"""Plotting functions for neurodsp.spectral."""

from itertools import repeat

import numpy as np
import matplotlib.pyplot as plt

from neurodsp.plts.style import style_plot
from neurodsp.plts.utils import check_ax, savefig

###################################################################################################
###################################################################################################

@savefig
@style_plot
def plot_power_spectra(freqs, powers, labels=None, colors=None, ax=None):
    """Plot power spectra.

    Parameters
    ----------
    freqs : 1d array or list of 1d array
        Frequency vector.
    powers : 1d array or list of 1d array
        Power values.
    labels : str or list of str, optional
        Labels for each time series.
    colors : str or list of str
        Colors to use to plot lines.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (6, 6))

    freqs = repeat(freqs) if isinstance(freqs, np.ndarray) else freqs
    powers = [powers] if isinstance(powers, np.ndarray) else powers

    if labels is not None:
        labels = [labels] if not isinstance(labels, list) else labels
    else:
        labels = repeat(labels)

    if colors is not None:
        colors = repeat(colors) if not isinstance(colors, list) else cycle(colors)

    for freq, power, label in zip(freqs, powers, labels):
        ax.loglog(freq, power, label=label)

    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('Power (V^2/Hz)')


@savefig
@style_plot
def plot_scv(freqs, scv, ax=None):
    """Plot the SCV.

    Parameters
    ----------
    freqs : 1d array
        Frequency vector.
    scv : 1d array
        Spectral coefficient of variation.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (5, 5))

    ax.loglog(freqs, scv)

    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('SCV')


@savefig
@style_plot
def plot_scv_rs_lines(freqs, scv_rs, ax=None):
    """Plot the SCV, from the resampling method.

    Parameters
    ----------
    freqs : 1d array
        Frequency vector.
    scv_rs :
        Spectral coefficient of variation, from resampling procedure.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (8, 8))

    ax.loglog(freqs, scv_rs, 'k', alpha=0.1)
    ax.loglog(freqs, np.mean(scv_rs, axis=1), lw=2)
    ax.loglog(freqs, len(freqs)*[1.])

    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('SCV')


@savefig
@style_plot
def plot_scv_rs_matrix(freqs, t_inds, scv_rs):
    """Plot the SCV, from the resampling method.

    Parameters
    ----------
    freqs : 1d array
        Frequency vector.
    t_inds : 1d array
        Time indices
    scv_rs : 1d array
        Spectral coefficient of variation, from resampling procedure.
    """

    fig, ax = plt.subplots(figsize=(10, 5))

    plt.imshow(np.log10(scv_rs), aspect='auto',
               extent=(t_inds[0], t_inds[-1], freqs[-1], freqs[0]))
    plt.colorbar(label='SCV')

    plt.xlabel('Time (s)')
    plt.ylabel('Frequency (Hz)')


@savefig
@style_plot
def plot_spectral_hist(freqs, power_bins, spectral_hist, spectrum_freqs=None, spectrum=None):
    """Plot the spectral histogram.

    Parameters
    ----------
    freqs : 1d array
        Frequencies over which the histogram is calculated.
    power_bins : 1d array
        Power bins within which histogram is aggregated.
    spectral_hist : 2d array
        Spectral histogram to be plotted.
    spectrum_freqs : 1d array, optional
        Frequency axis of the power spectrum to be plotted.
    spectrum : 1d array, optional
        Spectrum to be plotted over the histograms.
    """

    # Automatically scale figure height based on number of bins
    plt.figure(figsize=(8, 12 * len(power_bins) / len(freqs)))

    # Plot histogram intensity as image and automatically adjust aspect ratio
    plt.imshow(spectral_hist, extent=[freqs[0], freqs[-1], power_bins[0], power_bins[-1]], aspect='auto')
    plt.xlabel('Frequency (Hz)', fontsize=15)
    plt.ylabel('Log10 Power', fontsize=15)
    plt.colorbar(label='Probability')

    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Log10 Power')

    # If a PSD is provided, plot over the histogram data
    if spectrum is not None:
        plt_inds = np.logical_and(spectrum_freqs >= freqs[0], spectrum_freqs <= freqs[-1])
        plt.plot(spectrum_freqs[plt_inds], np.log10(spectrum[plt_inds]), color='w', alpha=0.8)



"""Utility functions for NeuroDSP simulations."""

import numpy as np

###################################################################################################
###################################################################################################

def create_times(n_seconds, fs, start_val=0.):
    """Create an array of time indices.

    Parameters
    ----------
    n_seconds : float
        Signal duration, in seconds.
    fs : float
        Signal sampling rate, in Hz.
    start_val : float, optional, default=0
        Starting value for the time definition.

    Returns
    -------
    1d array
        Time indices.
    """

    return np.arange(start_val, n_seconds, 1/fs)



"""Plotting functions for neurodsp.rhythm."""

import matplotlib.pyplot as plt

from neurodsp.plts.style import style_plot
from neurodsp.plts.utils import check_ax, savefig

###################################################################################################
###################################################################################################

@savefig
@style_plot
def plot_swm_pattern(pattern, ax=None):
    """Plot the resulting pattern from a sliding window matching analysis.

    Parameters
    ----------
    pattern : 1d array
        The resulting average pattern from applying sliding window matching.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (4, 4))

    plt.plot(pattern, 'k')

    plt.title('Average Pattern')
    plt.xlabel('Time (samples)')
    plt.ylabel('Voltage (a.u.)')


@savefig
@style_plot
def plot_lagged_coherence(freqs, lcs, ax=None):
    """Plot lagged coherence values across frequencies.

    Parameters
    ----------
    freqs : 1d array
        Vector of frequencies at which lagged coherence was computed.
    lcs : 1d array
        Lagged coherence values across the computed frequencies.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (6, 3))

    plt.plot(freqs, lcs, 'k.-')

    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Lagged Coherence')



"""Filter signals with IIR filters."""

from warnings import warn

from scipy.signal import butter, filtfilt

from neurodsp.utils import remove_nans, restore_nans
from neurodsp.utils.decorators import multidim
from neurodsp.filt.utils import compute_nyquist, compute_frequency_response
from neurodsp.filt.checks import check_filter_definition, check_filter_properties
from neurodsp.plts.filt import plot_frequency_response

###################################################################################################
###################################################################################################

@multidim
def filter_signal_iir(sig, fs, pass_type, f_range, butterworth_order,
                      print_transitions=False, plot_properties=False, return_filter=False):
    """Apply an IIR filter to a signal.

    Parameters
    ----------
    sig : 1d array
        Time series to be filtered.
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.
    butterworth_order : int
        Order of the butterworth filter, if using an IIR filter.
        See input 'N' in scipy.signal.butter.
    print_transitions : bool, optional, default: False
        If True, print out the transition and pass bandwidths.
    plot_properties : bool, optional, default: False
        If True, plot the properties of the filter, including frequency response and/or kernel.
    return_filter : bool, optional, default: False
        If True, return the filter coefficients of the IIR filter.

    Returns
    -------
    sig_filt : 1d array
        Filtered time series.
    filter_coefs : tuple of (1d array, 1d array)
        Filter coefficients of the IIR filter, as (b_vals, a_vals).
        Only returned if `return_filter` is True.
    """

    # Design filter
    b_vals, a_vals = design_iir_filter(fs, pass_type, f_range, butterworth_order)

    # Check filter properties: compute transition bandwidth & run checks
    check_filter_properties(b_vals, a_vals, fs, pass_type, f_range, verbose=print_transitions)

    # Remove any NaN on the edges of 'sig'
    sig, sig_nans = remove_nans(sig)

    # Apply filter
    sig_filt = filtfilt(b_vals, a_vals, sig)

    # Add NaN back on the edges of 'sig', if there were any at the beginning
    sig_filt = restore_nans(sig_filt, sig_nans)

    # Plot frequency response, if desired
    if plot_properties:
        f_db, db = compute_frequency_response(b_vals, a_vals, fs)
        plot_frequency_response(f_db, db)

    if return_filter:
        return sig_filt, (b_vals, a_vals)
    else:
        return sig_filt


def design_iir_filter(fs, pass_type, f_range, butterworth_order):
    """Design an IIR filter.

    Parameters
    ----------
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.
    butterworth_order : int
        Order of the butterworth filter, if using an IIR filter.
        See input 'N' in scipy.signal.butter.

    Returns
    -------
    b_vals : 1d array
        B value filter coefficients for an IIR filter.
    a_vals : 1d array
        A value filter coefficients for an IIR filter.
    """

    # Warn about only recommending IIR for bandstop
    if pass_type != 'bandstop':
        warn('IIR filters are not recommended other than for notch filters.')

    # Check filter definition
    f_lo, f_hi = check_filter_definition(pass_type, f_range)

    f_nyq = compute_nyquist(fs)
    if pass_type in ('bandpass', 'bandstop'):
        win = (f_lo / f_nyq, f_hi / f_nyq)
    elif pass_type == 'highpass':
        win = f_lo / f_nyq
    elif pass_type == 'lowpass':
        win = f_hi / f_nyq

    # Design filter
    b_vals, a_vals = butter(butterworth_order, win, pass_type)

    return b_vals, a_vals



"""Simulating time series, with aperiodic activity."""

import numpy as np
from scipy.stats import zscore

from neurodsp.filt import filter_signal, infer_passtype
from neurodsp.spectral import rotate_powerlaw
from neurodsp.utils.decorators import normalize
from neurodsp.sim.transients import sim_synaptic_kernel

###################################################################################################
###################################################################################################

@normalize
def sim_poisson_pop(n_seconds, fs, n_neurons=1000, firing_rate=2):
    """Simulates a poisson population.

    Parameters
    ----------
    n_seconds : float
        Simulation time, in seconds.
    fs : float
        Sampling rate of simulated signal, in Hz.
    n_neurons : int
        Number of neurons in the simulated population.
    firing_rate : type
        Firing rate of individual neurons in the population.

    Returns
    -------
    sig : 1d array
        Simulated population activity.

    Notes
    -----
    It is essentially white noise, but satisfies the Poisson property, i.e. mean(X) = var(X).

    The lambda parameter of the Poisson process (total rate) is determined as
    firing rate * number of neurons, i.e. summation of poisson processes is still
    a poisson processes.

    Note that the Gaussian approximation for a sum of Poisson processes is only
    a good approximation for large lambdas.
    """

    # Poisson population rate signal scales with # of neurons and individual rate
    lam = n_neurons * firing_rate

    # Variance is equal to the mean
    sig = np.random.normal(loc=lam, scale=lam**0.5, size=int(n_seconds * fs))

    # Enforce that sig is non-negative in cases of low firing rate
    sig[np.where(sig < 0.)] = 0.

    return sig


@normalize
def sim_synaptic_current(n_seconds, fs, n_neurons=1000, firing_rate=2,
                         tau_r=0, tau_d=0.01, t_ker=None):
    """Simulate a neural signal as synaptic current, which has 1/f characteristics with a knee.

    Parameters
    ----------
    n_seconds : float
        Simulation time, in seconds.
    fs : float
        Sampling rate of simulated signal, in Hz.
    n_neurons : int
        Number of neurons in the simulated population.
    firing_rate : float
        Firing rate of individual neurons in the population.
    tau_r : float
        Rise time of synaptic kernel, in seconds.
    tau_d : fload
        Decay time of synaptic kernel, in seconds.

    Returns
    -------
    sig : 1d array
        Simulated signal.

    Notes
    -----
    The resulting signal is most similar to unsigned intracellular current or conductance change.
    """

    # If not provided, compute t_ker as a function of decay time constant
    if t_ker is None:
        t_ker = 5. * tau_d

    # Simulate an extra bit because the convolution will snip it. Turn off normalization for this sig
    sig = sim_poisson_pop((n_seconds + t_ker), fs, n_neurons, firing_rate, mean=None, variance=None)
    ker = sim_synaptic_kernel(t_ker, fs, tau_r, tau_d)
    sig = np.convolve(sig, ker, 'valid')[:-1]

    return sig


@normalize
def sim_random_walk(n_seconds, fs, theta=1., mu=0., sigma=5.):
    """Simulate mean-reverting random walk, as an Ornstein-Uhlenbeck process.

    Parameters
    ----------
    n_seconds : float
        Simulation time, in seconds.
    fs : float
        Sampling rate of simulated signal, in Hz.
    theta : float
        Memory scale - larger theta = faster fluctuation.
    mu : float
        Mean.
    sigma : float
        Standard deviation.

    Returns
    -------
    sig: 1d array
        Simulated signal.

    Notes
    -----
    Discretized Ornstein-Uhlenbeck process:
    dx = theta*(x-mu)*dt + sigma*dWt
    dWt : increments of Wiener process, i.e. white noise
    theta : memory scale (higher = faster fluc)
    mu : mean
    sigma : std

    References
    ----------
    See for integral solution:
        https://en.wikipedia.org/wiki/Ornstein%E2%80%93Uhlenbeck_process#Solution
    """

    times = np.arange(0, n_seconds, 1 / fs)
    x0 = mu
    dt = times[1] - times[0]
    ws = np.random.normal(size=len(times))
    ex = np.exp(-theta * times)
    ws[0] = 0.
    sig = x0 * ex + mu * (1. - ex) + sigma * ex * \
        np.cumsum(np.exp(theta * times) * np.sqrt(dt) * ws)

    return sig


@normalize
def sim_powerlaw(n_seconds, fs, exponent=-2.0, f_range=None, **filter_kwargs):
    """Generate a power law time series with specified exponent by spectrally rotating white noise.

    Parameters
    ----------
    n_seconds : float
        Simulation time, in seconds.
    fs : float
        Sampling rate of simulated signal, in Hz.
    exponent : float
        Desired power-law exponent, of the form P(f)=f^exponent.
    f_range : list of [float, float] or None, optional
        Frequency range to filter simulated data, as [f_lo, f_hi], in Hz.
    **filter_kwargs : kwargs, optional
        Keyword arguments to pass to `filter_signal`.

    Returns
    -------
    sig: 1d array
        Time-series with the desired power-law exponent.
    """

    # NOTE: current hack to add an extra sample, to account for skipping f=0 below
    n_samples = int(n_seconds * fs)
    sig = np.random.randn(n_samples)

    # Compute the FFT
    fft_output = np.fft.fft(sig)
    freqs = np.fft.fftfreq(len(sig), 1. / fs)

    # Rotate spectrum and invert, zscore to normalize.
    #   Note: the delta exponent to be applied is divided by two, as
    #     the FFT output is in units of amplitude not power
    fft_output_rot = rotate_powerlaw(freqs, fft_output, -exponent/2)
    sig = zscore(np.real(np.fft.ifft(fft_output_rot)))

    if f_range is not None:
        filter_signal(sig, fs, infer_passtype(f_range), f_range, **filter_kwargs)

    return sig



"""Plotting functions."""

from .time_series import plot_time_series, plot_bursts
from .filt import plot_filter_properties, plot_frequency_response, plot_impulse_response
from .rhythm import plot_swm_pattern, plot_lagged_coherence
from .spectral import plot_power_spectra, plot_scv, plot_scv_rs_lines, plot_scv_rs_matrix, plot_spectral_hist



"""Checker functions for neurodsp.spectral."""

###################################################################################################
###################################################################################################

def check_spg_settings(fs, window, nperseg, noverlap):
    """Check settings used for calculating spectrogram."""

    # Set the nperseg, if not provided:
    if nperseg is None:

        # If the window is a string or tuple, defaults to 1 second of data
        if isinstance(window, (str, tuple)):
            nperseg = int(fs)
        # If the window is an array, defaults to window length
        else:
            nperseg = len(window)
    else:
        nperseg = int(nperseg)

    if noverlap is not None:
        noverlap = int(noverlap)

    return nperseg, noverlap



"""Filter signals with FIR filters."""

import numpy as np
from scipy.signal import firwin

from neurodsp.utils import remove_nans, restore_nans
from neurodsp.utils.decorators import multidim
from neurodsp.plts.filt import plot_filter_properties
from neurodsp.filt.utils import compute_nyquist, compute_frequency_response, remove_filter_edges
from neurodsp.filt.checks import check_filter_definition, check_filter_properties

###################################################################################################
###################################################################################################

@multidim
def filter_signal_fir(sig, fs, pass_type, f_range, n_cycles=3, n_seconds=None, remove_edges=True,
                      print_transitions=False, plot_properties=False, return_filter=False):
    """Apply an FIR filter to a signal.

    Parameters
    ----------
    sig : 1d array
        Time series to be filtered.
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.
    n_cycles : float, optional, default: 3
        Length of filter, in number of cycles, defined at the 'f_lo' frequency.
        This parameter is overwritten by `n_seconds`, if provided.
    n_seconds : float, optional
        Length of filter, in seconds. This parameter overwrites `n_cycles`.
    remove_edges : bool, optional
        If True, replace samples within half the kernel length to be np.nan.
    print_transitions : bool, optional, default: False
        If True, print out the transition and pass bandwidths.
    plot_properties : bool, optional, default: False
        If True, plot the properties of the filter, including frequency response and/or kernel.
    return_filter : bool, optional, default: False
        If True, return the filter coefficients of the FIR filter.

    Returns
    -------
    sig_filt : 1d array
        Filtered time series.
    filter_coefs : 1d array
        Filter coefficients of the FIR filter. Only returned if `return_filter` is True.
    """

    # Design filter
    filter_coefs = design_fir_filter(len(sig), fs, pass_type, f_range, n_cycles, n_seconds)

    # Check filter properties: compute transition bandwidth & run checks
    check_filter_properties(filter_coefs, 1, fs, pass_type, f_range, verbose=print_transitions)

    # Remove any NaN on the edges of 'sig'
    sig, sig_nans = remove_nans(sig)

    # Apply filter
    sig_filt = np.convolve(filter_coefs, sig, 'same')

    # Remove edge artifacts
    if remove_edges:
        sig_filt = remove_filter_edges(sig_filt, len(filter_coefs))

    # Add NaN back on the edges of 'sig', if there were any at the beginning
    sig_filt = restore_nans(sig_filt, sig_nans)

    # Plot filter properties, if specified
    if plot_properties:
        f_db, db = compute_frequency_response(filter_coefs, 1, fs)
        plot_filter_properties(f_db, db, fs, filter_coefs)

    if return_filter:
        return sig_filt, filter_coefs
    else:
        return sig_filt


def design_fir_filter(sig_length, fs, pass_type, f_range, n_cycles=3, n_seconds=None):
    """Design an FIR filter.

    Parameters
    ----------
    sig_length : int
        The length of the signal to be filtered.
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.
    n_cycles : float, optional, default: 3
        Length of filter, in number of cycles, defined at the 'f_lo' frequency.
        This parameter is overwritten by `n_seconds`, if provided.
    n_seconds : float, optional
        Length of filter, in seconds. This parameter overwrites `n_cycles`.

    Returns
    -------
    filter_coefs : 1d array
        The filter coefficients for an FIR filter.
    """

    # Check filter definition
    f_lo, f_hi = check_filter_definition(pass_type, f_range)
    filt_len = compute_filt_len(sig_length, fs, pass_type, f_lo, f_hi, n_cycles, n_seconds)

    f_nyq = compute_nyquist(fs)
    if pass_type == 'bandpass':
        filter_coefs = firwin(filt_len, (f_lo, f_hi), pass_zero=False, nyq=f_nyq)
    elif pass_type == 'bandstop':
        filter_coefs = firwin(filt_len, (f_lo, f_hi), nyq=f_nyq)
    elif pass_type == 'highpass':
        filter_coefs = firwin(filt_len, f_lo, pass_zero=False, nyq=f_nyq)
    elif pass_type == 'lowpass':
        filter_coefs = firwin(filt_len, f_hi, nyq=f_nyq)

    return filter_coefs


def compute_filt_len(sig_length, fs, pass_type, f_lo, f_hi, n_cycles, n_seconds):
    """Calculate and check the filter length for an FIR signal with specified parameters.

    Parameters
    ----------
    sig_length : int
        The length of the signal to be filtered.
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply.
    f_lo : float or None
        The lower frequency range of the filter, specifying the highpass frequency, if specified.
    f_hi : float or None
        The higher frequency range of the filter, specifying the lowpass frequency, if specified.
    n_cycles : float, optional, default: 3
        Length of filter, in number of cycles, defined at the 'f_lo' frequency.
    n_seconds : float, optional
        Length of filter, in seconds.

    Returns
    -------
    filt_len : int
        The length of the specified filter.
    """

    # Compute filter length if specified in seconds
    if n_seconds is not None:
        filt_len = fs * n_seconds
    else:
        if pass_type == 'lowpass':
            filt_len = fs * n_cycles / f_hi
        else:
            filt_len = fs * n_cycles / f_lo

    # Typecast filter length to an integer, rounding up
    filt_len = int(np.ceil(filt_len))

    # Force filter length to be odd
    if filt_len % 2 == 0:
        filt_len = filt_len + 1

    # Raise an error if the filter is longer than the signal
    if filt_len >= sig_length:
        raise ValueError(
            'The designed filter (length: {:d}) is longer than the signal '\
            '(length: {:d}). The filter needs to be shortened by decreasing '\
            'the n_cycles or n_seconds parameter. However, this will decrease '\
            'the frequency resolution of the filter.'.format(filt_len, sig_length))

    return filt_len



"""Simulating time series, with periodic activity."""

import numpy as np
from numpy.random import rand, randn, randint
import pandas as pd

from neurodsp.utils.decorators import normalize
from neurodsp.sim.transients import sim_osc_cycle

###################################################################################################
###################################################################################################

@normalize
def sim_oscillation(n_seconds, fs, freq, cycle='sine', **cycle_params):
    """Simulate an oscillation.

    Parameters
    ----------
    n_seconds : float
        Signal duration, in seconds.
    fs : float
        Signal sampling rate, in Hz.
    freq : float
        Oscillation frequency.
    cycle : {'sine', 'asine', 'sawtooth', 'gaussian', 'exp', '2exp'}
        What type of oscillation cycle to simulate.
        See `sim_osc_cycle` for details on cycle types and parameters.
    **cycle_params
        Parameters for the simulated oscillation cycle.

    Returns
    -------
    osc : 1d array
        Oscillating time series.
    """

    n_cycles = int(np.ceil(n_seconds * freq))
    n_seconds_cycle = int(np.ceil(fs / freq)) / fs

    osc_cycle = sim_osc_cycle(n_seconds_cycle, fs, cycle, **cycle_params)
    osc = np.tile(osc_cycle, n_cycles)

    return osc


@normalize
def sim_bursty_oscillation(n_seconds, fs, freq, enter_burst=.2, leave_burst=.2,
                           cycle='sine', **cycle_params):
    """Simulate a bursty oscillation.

    Parameters
    ----------
    n_seconds : float
        Simulation time, in seconds.
    fs : float
        Sampling rate of simulated signal, in Hz
    freq : float
        Oscillation frequency, in Hz.
    enter_burst : float
        Probability of a cycle being oscillating given the last cycle is not oscillating.
    leave_burst : float
        Probability of a cycle not being oscillating given the last cycle is oscillating.
    cycle : {'sine', 'asine', 'sawtooth', 'gaussian', 'exp', '2exp'}
        What type of oscillation cycle to simulate.
        See `sim_osc_cycle` for details on cycle types and parameters.
    **cycle_params
        Parameters for the simulated oscillation cycle.

    Returns
    -------
    sig : 1d array
        Bursty oscillation.

    Notes
    -----
    * This function takes a 'tiled' approach to simulating cycles, with evenly spaced
    and consistent cycles across the whole signal, that are either oscillating or not.
    * If the cycle length does not fit evenly into the simulated data length,
    then the last few cycle will be non-oscillating.
    """

    # Determine number of samples & cycles
    n_samples = int(n_seconds * fs)
    n_seconds_cycle = (1/freq * fs)/fs

    # Make a single cycle of an oscillation
    osc_cycle = sim_osc_cycle(n_seconds_cycle, fs, cycle, **cycle_params)
    n_samples_cycle = len(osc_cycle)
    n_cycles = int(np.floor(n_samples / n_samples_cycle))

    # Determine which periods will be oscillating
    is_oscillating = _make_is_osc(n_cycles, enter_burst, leave_burst)

    # Fill in the signal with cycle oscillations, for all bursting cycles
    sig = np.zeros([n_samples])
    for is_osc, cycle_ind in zip(is_oscillating, range(0, n_samples, n_samples_cycle)):
        if is_osc:
            sig[cycle_ind:cycle_ind+n_samples_cycle] = osc_cycle

    return sig

###################################################################################################
###################################################################################################

def _make_is_osc(n_cycles, enter_burst, leave_burst):
    """Create a vector describing if each cycle is oscillating, for bursting oscillations."""

    is_oscillating = [None] * (n_cycles)
    is_oscillating[0] = False

    for ii in range(1, n_cycles):

        rand_num = rand()

        if is_oscillating[ii-1]:
            is_oscillating[ii] = rand_num > leave_burst
        else:
            is_oscillating[ii] = rand_num < enter_burst

    return is_oscillating



"""Test spectral plots."""

import numpy as np

from neurodsp.spectral.spectral import compute_spectral_hist
from neurodsp.tests.utils import plot_test

from neurodsp.plts.spectral import *

###################################################################################################
###################################################################################################

@plot_test
def test_plot_power_spectra():

    freqs, powers = np.array([1, 2, 3, 4]), np.array([10, 20, 30, 40])
    plot_power_spectra(freqs, powers)
    plot_power_spectra([freqs, freqs], [powers, powers])

@plot_test
def test_plot_scv():

    freqs, scv = np.array([1, 2, 3, 4]), np.array([10, 20, 30, 40])
    plot_scv(freqs, scv)

#@plot_test
def test_plot_scv_rs_lines():

    pass

#@plot_test
def test_plot_scv_rs_matrix():
    pass

@plot_test
def test_plot_spectral_hist(tsig):

    freqs, power_bins, spectral_hist = compute_spectral_hist(tsig, fs=1000)
    plot_spectral_hist(freqs, power_bins, spectral_hist)



"""Utility function for neurodsp.spectral."""

import numpy as np

###################################################################################################
###################################################################################################

def trim_spectrum(freqs, power_spectra, f_range):
    """Extract frequency range of interest from power spectra.

    Parameters
    ----------
    freqs : 1d array
        Frequency values for the PSD.
    power_spectra : 1d or 2d array
        Power spectral density values.
    f_range: list of [float, float]
        Frequency range to restrict to.

    Returns
    -------
    freqs_ext : 1d array
        Extracted frequency values for the power spectrum.
    power_spectra_ext : 1d or 2d array
        Extracted power spectral density values.

    Notes
    -----
    This function extracts frequency ranges >= f_low and <= f_high.
    It does not round to below or above f_low and f_high, respectively.
    """

    # Create mask to index only requested frequencies
    f_mask = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])

    # Restrict freqs & psd to requested range. The if/else is to cover both 1d or 2d arrays
    freqs_ext = freqs[f_mask]
    power_spectra_ext = power_spectra[f_mask] if power_spectra.ndim == 1 \
        else power_spectra[:, f_mask]

    return freqs_ext, power_spectra_ext


def rotate_powerlaw(freqs, spectrum, delta_exponent, f_rotation=1):
    """Change the power law exponent of a power spectrum about an axis frequency.

    Parameters
    ----------
    freqs : 1d array
        Frequency axis of input spectrum, in Hz. Must be same length as spectrum.
    spectrum : 1d array
        Power spectrum to be rotated.
    delta_exponent : float
        Change in power law exponent to be applied.
        Positive is counterclockwise rotation (flatten), negative is clockwise rotation (steepen).
    f_rotation : float, optional, default=1
        Axis of rotation frequency, in Hz, such that power at that frequency is unchanged
        by the rotation. Only matters if not further normalizing signal variance.

    Returns
    -------
    rot_spectrum : 1d array
        Rotated spectrum.
    """

    if freqs[0] == 0:
        skipped_zero = True
        f_0, p_0 = freqs[0], spectrum[0]
        freqs, spectrum = freqs[1:], spectrum[1:]
    else:
        skipped_zero = False

    mask = (np.abs(freqs) / f_rotation)**-delta_exponent
    rotated_spectrum = mask * spectrum

    if skipped_zero:
        freqs = np.insert(freqs, 0, f_0),
        rotated_spectrum =  np.insert(rotated_spectrum, 0, p_0)

    return rotated_spectrum



"""Tests for filter utilities."""

from neurodsp.filt.utils import *
from neurodsp.filt.fir import design_fir_filter, compute_filt_len

###################################################################################################
###################################################################################################

def test_infer_passtype():

    assert infer_passtype((None, 1)) == 'lowpass'
    assert infer_passtype((1, None)) == 'highpass'
    assert infer_passtype((1, 2)) == 'bandpass'

def test_compute_frequency_response():

    filter_coefs = design_fir_filter(1000, 500, 'bandpass', (8, 12))
    f_db, db = compute_frequency_response(filter_coefs, 1, 500)
    assert True

def test_compute_pass_band():

    fs = 500
    assert compute_pass_band(fs, 'bandpass', (4, 8)) == 4.
    assert compute_pass_band(fs, 'lowpass', 20) == 20.
    assert compute_pass_band(fs, 'highpass', 5) == compute_nyquist(fs) - 5

def test_compute_transition_band():

    filter_coefs = design_fir_filter(1000, 500, 'bandpass', (8, 12))
    f_db, db = compute_frequency_response(filter_coefs, 1, 500)
    trans_band = compute_transition_band(f_db, db)
    assert True

def test_compute_nyquist():

    assert compute_nyquist(100.) == 50.
    assert compute_nyquist(256) == 128.

def test_remove_filter_edges():

    # Get the length for a possible filter & calc # of values should be dropped for it
    sig_len = 1000
    sig = np.ones([1, sig_len])
    filt_len = compute_filt_len(sig_len, 500, 'bandpass', 4, 8, 3, None)
    n_rmv = int(np.ceil(filt_len / 2))

    dropped_sig = remove_filter_edges(sig, filt_len)

    assert np.all(np.isnan(dropped_sig[:n_rmv]))
    assert np.all(np.isnan(dropped_sig[-n_rmv:]))
    assert np.all(~np.isnan(dropped_sig[n_rmv:-n_rmv]))



"""Compute spectral measures, including power spectra and spectral variance.

Notes
-----
Mostly relies on scipy.signal.spectrogram and numpy.fft
https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.spectrogram.html
"""

import numpy as np
from scipy.signal import spectrogram, medfilt

from neurodsp.utils import discard_outliers
from neurodsp.utils.core import get_avg_func
from neurodsp.utils.decorators import multidim
from neurodsp.timefrequency.wavelets import morlet_transform
from neurodsp.spectral.utils import trim_spectrum
from neurodsp.spectral.checks import check_spg_settings

###################################################################################################
###################################################################################################

@multidim
def compute_spectrum(sig, fs, method='welch', avg_type='mean', **kwargs):
    """Estimate the power spectral density (PSD) of a time series.

    Parameters
    -----------
    sig : 1d or 2d array
        Time series of measurement values.
    fs : float
        Sampling rate, in Hz.
    method : {'welch', 'wavelet', 'medfilt'}
        Method to use to estimate the power spectrum.
    avg_type : {'mean', 'median'}, optional
        If relevant, the method to average across windows to create the spectrum.
    **kwargs
        Keyword arguments to pass through to function that calculates the spectrum.

    Returns
    -------
    freqs : 1d array
        Array of sample frequencies.
    spectrum : 1d or 2d array
        Power spectral density.
    """

    if method not in ('welch', 'medfilt', 'wavelet'):
        raise ValueError('Unknown power spectrum method: %s' % method)

    if method == 'welch':
        return compute_spectrum_welch(sig, fs, avg_type=avg_type, **kwargs)

    elif method == 'wavelet':
        return compute_spectrum_wavelet(sig, fs, avg_type=avg_type, **kwargs)

    elif method == 'medfilt':
        return compute_spectrum_medfilt(sig, fs, **kwargs)


@multidim
def compute_spectrum_wavelet(sig, fs, freqs, avg_type='mean', **kwargs):
    """Estimate the power spectral densitry using wavelets.

    Parameters
    ----------
    sig : 1d or 2d array
        Time series of measurement values.
    fs : float
        Sampling rate, in Hz.
    avg_type : {'mean', 'median'}, optional
        Method to average across the windows.
    **kwargs
        Optional inputs for using wavelets.

    Returns
    -------
    freqs : 1d array
        Array of sample frequencies.
    spectrum : 1d or 2d array
        Power spectral density.
    """

    mwt = morlet_transform(sig, fs, freqs, **kwargs)
    spectrum = get_avg_func(avg_type)(mwt, axis=0)

    return freqs, spectrum


@multidim
def compute_spectrum_welch(sig, fs, avg_type='mean', window='hann',
                           nperseg=None, noverlap=None,
                           f_range=None, outlier_pct=None):
    """Estimate the power spectral density using Welch's method.

    Parameters
    -----------
    sig : 1d or 2d array
        Time series of measurement values.
    fs : float
        Sampling rate, in Hz.
    avg_type : {'mean', 'median'}, optional
        Method to average across the windows:

        * 'mean' is the same as Welch's method, taking the mean across FFT windows.
        * 'median' uses median across FFT windows instead of the mean, to minimize outlier effect.
    window : str or tuple or array_like, optional, default: 'hann'
        Desired window to use. Defaults to a Hann window.
        See scipy.signal.get_window for a list of windows and required parameters.
        If array_like, the array will be used as the window and its length must be nperseg.
    nperseg : int, optional
        Length of each segment, in number of samples.
        If None, and window is str or tuple, is set to 1 second of data.
        If None, and window is array_like, is set to the length of the window.
    noverlap : int, optional
        Number of points to overlap between segments.
        If None, noverlap = nperseg // 8.
    f_range : list of [float, float] optional
        Frequency range to sub-select from the power spectrum.
    outlier_pct : float, optional
        Percentage of the windows with the lowest and highest total log power to discard.
        Must be between 0 and 100.

    Returns
    -------
    freqs : 1d array
        Array of sample frequencies.
    spectrum : 1d or 2d array
        Power spectral density.
    """

    # Calculate the short time fourier transform with signal.spectrogram
    nperseg, noverlap = check_spg_settings(fs, window, nperseg, noverlap)
    freqs, _, spg = spectrogram(sig, fs, window, nperseg, noverlap)

    # Pad data to 2D
    if len(sig.shape) == 1:
        sig = sig[np.newaxis :]

    # Throw out outliers if indicated
    if outlier_pct is not None:
        spg = discard_outliers(spg, outlier_pct)

    # Average across windows
    spectrum = get_avg_func(avg_type)(spg, axis=-1)

    # Trim spectrum, if requested
    if f_range:
        freqs, spectrum = trim_spectrum(freqs, spectrum, f_range)

    return freqs, spectrum


@multidim
def compute_spectrum_medfilt(sig, fs, filt_len=1., f_range=None):
    """Estimate the power spectral densitry as a smoothed FFT.

    Parameters
    ----------
    sig : 1d or 2d array
        Time series of measurement values.
    fs : float
        Sampling rate, in Hz.
    filt_len : float, optional, default: 1.
        Length of the median filter, in Hz.
    f_range : list of [float, float] optional
        Frequency range to sub-select from the power spectrum.

    Returns
    -------
    freqs : 1d array
        Array of sample frequencies.
    spectrum : 1d or 2d array
        Power spectral density.
    """

    # Take the positive half of the spectrum since it's symmetrical
    ft = np.fft.fft(sig)[:int(np.ceil(len(sig) / 2.))]
    freqs = np.fft.fftfreq(len(sig), 1. / fs)[:int(np.ceil(len(sig) / 2.))]  # get freq axis

    # Convert median filter length from Hz to samples
    filt_len_samp = int(int(filt_len / (freqs[1] - freqs[0])) / 2 * 2 + 1)
    spectrum = medfilt(np.abs(ft)**2. / (fs * len(sig)), filt_len_samp)

    if f_range:
        freqs, spectrum = trim_spectrum(freqs, spectrum, f_range)

    return freqs, spectrum


@multidim
def compute_scv(sig, fs, window='hann', nperseg=None, noverlap=0, outlier_pct=None):
    """Compute the spectral coefficient of variation (SCV) at each frequency.

    Parameters
    -----------
    sig : 1d array
        Time series of measurement values.
    fs : float
        Sampling rate, in Hz.
    window : str or tuple or array_like, optional, default='hann'
        Desired window to use. Defaults to a Hann window.
        See scipy.signal.get_window for a list of windows and required parameters.
        If array_like, the array will be used as the window and its length must be nperseg.
    nperseg : int, optional
        Length of each segment, in number of samples.
        If None, and window is str or tuple, is set to 1 second of data.
        If None, and window is array_like, is set to the length of the window.
    noverlap : int, optional, default: 0
        Number of points to overlap between segments.
    outlier_pct : float, optional
        Percentage of the windows with the lowest and highest total log power to discard.
        Must be between 0 and 100.

    Returns
    -------
    freqs : 1d array
        Array of sample frequencies.
    scv : 1d array
        Spectral coefficient of variation.

    Notes
    -----
    White noise should have a SCV of 1 at all frequencies.
    """

    # Compute spectrogram of data
    nperseg, noverlap = check_spg_settings(fs, window, nperseg, noverlap)
    freqs, _, spg = spectrogram(sig, fs, window, nperseg, noverlap)

    if outlier_pct is not None:
        spg = discard_outliers(spg, outlier_pct)

    scv = np.std(spg, axis=-1) / np.mean(spg, axis=-1)

    return freqs, scv


@multidim
def compute_scv_rs(sig, fs, window='hann', nperseg=None, noverlap=0,
                   method='bootstrap', rs_params=None):
    """Resampled version of scv: instead of a single estimate of mean and standard deviation,
    the spectrogram is resampled, either randomly (bootstrap) or time-stepped (rolling).

    Parameters
    -----------
    sig : 1d array
        Time series of measurement values.
    fs : float
        Sampling rate, in Hz.
    window : str or tuple or array_like, optional, default='hann'
        Desired window to use. Defaults to a Hann window.
        See scipy.signal.get_window for a list of windows and required parameters.
        If array_like, the array will be used as the window and its length must be nperseg.
    nperseg : int, optional
        Length of each segment, in number of samples.
        If None, and window is str or tuple, is set to 1 second of data.
        If None, and window is array_like, is set to the length of the window.
    noverlap : int, optional, default: 0
        Number of points to overlap between segments.
    method : {'bootstrap', 'rolling'}, optional
        Method of resampling.
        'bootstrap' randomly samples a subset of the spectrogram repeatedly.
        'rolling' takes the rolling window scv.
    rs_params : tuple, (int, int), optional
        Parameters for resampling algorithm, depending on the method used.
        If 'bootstrap', rs_params = (nslices, ndraws), defaults to (10% of slices, 100 draws).
        This specifies the number of slices per draw, and number of random draws.
        If 'rolling', rs_params = (nslices, nsteps), defaults to (10, 5).
        This specifies the number of slices per draw, and number of slices to step forward.

    Returns
    -------
    freqs : 1d array
        Array of sample frequencies.
    t_inds : 1d array or None
        Array of time indices, for 'rolling' resampling. If 'bootstrap', t_inds = None.
    scv_rs : 2d array
        Resampled spectral coefficient of variation.
    """

    # Compute spectrogram of data
    nperseg, noverlap = check_spg_settings(fs, window, nperseg, noverlap)
    freqs, ts, spg = spectrogram(sig, fs, window, nperseg, noverlap)

    if method == 'bootstrap':

        # Params: number of slices of STFT to compute SCV over & number of draws
        #   Defaults to draw 1/10 of STFT slices, 100 draws
        if rs_params is None:
            rs_params = (int(spg.shape[1] / 10.), 100)

        nslices, ndraws = rs_params
        scv_rs = np.zeros((len(freqs), ndraws))

        # Repeated subsampling of spectrogram randomly, with replacement between draws
        for draw in range(ndraws):
            idx = np.random.choice(spg.shape[1], size=nslices, replace=False)
            scv_rs[:, draw] = np.std(
                spg[:, idx], axis=-1) / np.mean(spg[:, idx], axis=-1)

        t_inds = None  # no time component, return nothing

    elif method == 'rolling':

        # Params: number of slices of STFT to compute SCV over & number of slices to roll forward
        #   Defaults to 10 STFT slices, move forward by 5 slices
        if rs_params is None:
            rs_params = (10, 5)

        nslices, nsteps = rs_params
        outlen = int(np.ceil((spg.shape[1] - nslices) / float(nsteps))) + 1
        scv_rs = np.zeros((len(freqs), outlen))
        for ind in range(outlen):
            curblock = spg[:, nsteps * ind:nslices + nsteps * ind]
            scv_rs[:, ind] = np.std(
                curblock, axis=-1) / np.mean(curblock, axis=-1)

        # Grab the time indices from the spectrogram
        t_inds = ts[0::nsteps]

    else:
        raise ValueError('Unknown resampling method: %s' % method)

    return freqs, t_inds, scv_rs


@multidim
def compute_spectral_hist(sig, fs, window='hann', nperseg=None, noverlap=None,
                          nbins=50, f_range=[0., 100.], cut_pct=[0., 100.]):
    """Compute the distribution of log10 power at each frequency from the signal spectrogram.

    Parameters
    -----------
    sig : 1d array
        Time series of measurement values.
    fs : float
        Sampling rate, in Hz.
    window : str or tuple or array_like, optional, default='hann'
        Desired window to use. Defaults to a Hann window.
        See scipy.signal.get_window for a list of windows and required parameters.
        If array_like, the array will be used as the window and its length must be nperseg.
    nperseg : int, optional
        Length of each segment, in number of samples.
        If None, and window is str or tuple, is set to 1 second of data.
        If None, and window is array_like, is set to the length of the window.
    noverlap : int, optional
        Number of points to overlap between segments. If None, noverlap = nperseg // 2.
    nbins : int, optional, default: 50
        Number of histogram bins to use.
    f_range : list of [float, float], optional, default: [0, 100]
        Frequency range of the spectrogram to compute the histograms, as [start, end], in Hz.
    cut_pct : list of [float, float], optional, default: [0, 100]
        Power percentile at which to draw the lower and upper bin limits, as [low, high], in Hz.

    Returns
    -------
    freqs : 1d array
        Array of frequencies.
    power_bins : 1d array
        Histogram bins used to compute the distribution.
    spectral_hist : 2d array
        Power distribution at every frequency, nbins x fs 2D matrix.

    Notes
    -----
    The histogram bins are the same for every frequency, thus evenly spacing the global min and max power.
    """

    # Compute spectrogram of data
    nperseg, noverlap = check_spg_settings(fs, window, nperseg, noverlap)
    freqs, _, spg = spectrogram(sig, fs, window, nperseg, noverlap, return_onesided=True)

    # Get log10 power & limit to frequency range of interest before binning
    ps = np.transpose(np.log10(spg))
    freqs, ps = trim_spectrum(freqs, ps, f_range)

    # Prepare bins for power - min and max of bins determined by power cutoff percentage
    power_min, power_max = np.percentile(np.ndarray.flatten(ps), cut_pct)
    power_bins = np.linspace(power_min, power_max, nbins + 1)

    # Compute histogram of power for each frequency
    spectral_hist = np.zeros((len(ps[0]), nbins))
    for ind in range(len(ps[0])):
        spectral_hist[ind], _ = np.histogram(ps[:, ind], power_bins)
        spectral_hist[ind] = spectral_hist[ind] / sum(spectral_hist[ind])

    # Flip output for more sensible plotting direction
    spectral_hist = np.transpose(spectral_hist)
    spectral_hist = np.flipud(spectral_hist)

    return freqs, power_bins, spectral_hist



"""Test functions for timefrequency hilbert analyses."""

from neurodsp.timefrequency.hilbert import *

###################################################################################################
###################################################################################################

def test_robust_hilbert():

    # Generate a signal with NaNs
    fs, n_points, n_nans = 100, 1000, 10
    sig = np.random.randn(n_points)
    sig[0:n_nans] = np.nan

    # Check has correct number of nans (not all nan), without increase_n
    hilb_sig = robust_hilbert(sig)
    assert sum(np.isnan(hilb_sig)) == n_nans

    # Check has correct number of nans (not all nan), with increase_n
    hilb_sig = robust_hilbert(sig, True)
    assert sum(np.isnan(hilb_sig)) == n_nans

def test_phase_by_time(tsig):

    out = phase_by_time(tsig, 500, (8, 12))
    assert True

def test_amp_by_time(tsig):

    out = amp_by_time(tsig, 500, (8, 12))
    assert True

def test_freq_by_time(tsig):

    out = freq_by_time(tsig, 500, (8, 12))
    assert True



"""Utilities for accessing the simulation functions and retrieving information about them."""

from inspect import getmembers, isfunction

from neurodsp.sim import periodic, aperiodic, transients, combined

###################################################################################################
###################################################################################################

SIM_MODULES = ['periodic', 'aperiodic', 'transients', 'combined']

def get_sim_funcs(module_name):
    """Get the available sim functions from a specified sub-module.

    Parameters
    ----------
    module_name : {'periodic', 'aperiodic', 'transients', 'combined'}
        Simulation sub-module to get sim functions from.

    Returns
    -------
    funcs : dictionary
        A dictionary containing the available sim functions from the requested sub-module.
    """

    if module_name in SIM_MODULES:
        module = eval(module_name)
    else:
        raise ValueError('Requested sim module not understood.')

    funcs = {name : func for name, func in getmembers(module, isfunction) \
        if name[0:4] == 'sim_' and func.__module__.split('.')[-1] == module.__name__.split('.')[-1]}

    return funcs


def get_sim_names(module_name):
    """Get the names of the available sim functions from a specified sub-module.

    Parameters
    ----------
    module_name : {'periodic', 'aperiodic', 'transients', 'combined'}
        Simulation sub-module to get sim functions from.

    Returns
    -------
    list of str
        The names of the available functions in the requested sub-module.
    """

    return list(get_sim_funcs(module_name).keys())


def get_sim_func(function_name):
    """Get a specified sim function.

    Parameters
    ----------
    function_name : str
        Name of the sim function to retrieve.

    Returns
    -------
    func : executable
        Requested sim function.
    """

    for module in SIM_MODULES:
        try:
            func = get_sim_funcs(module)[function_name]
            break
        except KeyError:
            continue

    else:
        raise ValueError('Requested simulation function not found.') from None

    return func



"""Filter time series."""

from warnings import warn

from neurodsp.utils.decorators import multidim
from neurodsp.filt.fir import filter_signal_fir
from neurodsp.filt.iir import filter_signal_iir

###################################################################################################
###################################################################################################

@multidim
def filter_signal(sig, fs, pass_type, f_range, filt_type='fir',
                  n_cycles=3, n_seconds=None, remove_edges=True, butterworth_order=None,
                  print_transitions=False, plot_properties=False, return_filter=False):
    """Apply a bandpass, bandstop, highpass, or lowpass filter to a neural signal.

    Parameters
    ----------
    sig : 1d array
        Time series to be filtered.
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.
    n_cycles : float, optional, default: 3
        Length of filter, in number of cycles, at the 'f_lo' frequency, if using an FIR filter.
        This parameter is overwritten by `n_seconds`, if provided.
    n_seconds : float, optional
        Length of filter, in seconds, if using an FIR filter.
        This parameter overwrites `n_cycles`.
    filt_type : {'fir', 'iir'}, optional
        Whether to use an FIR or IIR filter.
        The only IIR filter offered is a butterworth filter.
    remove_edges : bool, optional, default: True
        If True, replace samples within half the kernel length to be np.nan.
        Only used for FIR filters.
    butterworth_order : int, optional
        Order of the butterworth filter, if using an IIR filter.
        See input 'N' in scipy.signal.butter.
    print_transitions : bool, optional, default: True
        If True, print out the transition and pass bandwidths.
    plot_properties : bool, optional, default: False
        If True, plot the properties of the filter, including frequency response and/or kernel.
    return_filter : bool, optional, default: False
        If True, return the filter coefficients.

    Returns
    -------
    sig_filt : 1d array
        Filtered time series.
    kernel : 1d array or tuple of (1d array, 1d array)
        Filter coefficients. Only returned if `return_filter` is True.
    """

    if filt_type == 'fir':
        return filter_signal_fir(sig, fs, pass_type, f_range, n_cycles, n_seconds,
                                 remove_edges, print_transitions,
                                 plot_properties, return_filter)
    elif filt_type == 'iir':
        _iir_checks(n_seconds, butterworth_order, remove_edges)
        return filter_signal_iir(sig, fs, pass_type, f_range, butterworth_order,
                                 print_transitions, plot_properties,
                                 return_filter)
    else:
        raise ValueError('Filter type not understood.')


def _iir_checks(n_seconds, butterworth_order, remove_edges):
    """Checks for using an IIR filter if called from the general filter function."""

    # Check inputs for IIR filters
    if n_seconds is not None:
        raise ValueError('n_seconds should not be defined for an IIR filter.')
    if butterworth_order is None:
        raise ValueError('butterworth_order must be defined when using an IIR filter.')
    if remove_edges:
        warn('Edge artifacts are not removed when using an IIR filter.')



"""Test filtering functions."""

from pytest import raises, warns

from neurodsp.filt.filter import *
from neurodsp.filt.filter import _iir_checks

###################################################################################################
###################################################################################################

def test_filter_signal(tsig):

    out = filter_signal(tsig, 500, 'bandpass', (8, 12), filt_type='fir')
    out = filter_signal(tsig, 500, 'bandpass', (8, 12), filt_type='iir', butterworth_order=3)
    with raises(ValueError):
        out = filter_signal(tsig, 500, 'bandpass', (8, 12), filt_type='bad')
    assert True

def test_iir_checks():

    # Check catch for having n_seconds defined
    with raises(ValueError):
        _iir_checks(1, 3, None)
    # Check catch for not having butterworth_order defined
    with raises(ValueError):
        _iir_checks(None, None, None)
    # Check catch for having remove_edges defined
    with warns(UserWarning):
        _iir_checks(None, 3, True)



"""The dual threshold algorithm for detecting oscillatory bursts in a neural signal."""

import numpy as np

from neurodsp.utils.core import get_avg_func
from neurodsp.utils.decorators import multidim
from neurodsp.timefrequency.hilbert import amp_by_time

###################################################################################################
###################################################################################################

@multidim
def detect_bursts_dual_threshold(sig, fs, f_range, dual_thresh, min_cycles=3,
                                 avg_type='median', magnitude_type='amplitude',
                                 **filter_kwargs):
    """Detect bursts in a neural signal using the dual threshold algorithm.

    Parameters
    ----------
    sig : 1d array
        Voltage time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of (float, float)
        Frequency range, in Hz, for narrowband signal of interest.
    dual_thresh : tuple of (float, float)
        Low and high threshold values for burst detection.
        Units are normalized by the average signal magnitude.
    min_cycles : float, optional, default=3
        Minimum burst duration in terms of number of cycles of f_range[0].
    avg_type : {'median', 'mean'}, optional
        Averaging method to use to normalize the magnitude that is used for thresholding.
    magnitude_type : {'amplitude', 'power'}, optional
        Metric of magnitude used for thresholding.
    **filter_kwargs
        Keyword parameters to pass to `filter_signal`.

    Returns
    -------
    is_burst : 1d array
        Boolean indication of where bursts are present in the input signal.
        True indicates that a burst was detected at that sample, otherwise False.
    """

    if len(dual_thresh) != 2:
        raise ValueError("Invalid number of elements in 'dual_thresh' parameter")

    # Compute amplitude time series
    sig_magnitude = amp_by_time(sig, fs, f_range, remove_edges=False, **filter_kwargs)

    # Set magnitude as power or amplitude: square if power, leave as is if amplitude
    if magnitude_type not in ['amplitude', 'power']:
        raise ValueError("Invalid input for 'magnitude_type'")
    if magnitude_type == 'power':
        sig_magnitude = sig_magnitude**2

    # Calculate normalized magnitude
    sig_magnitude = sig_magnitude / get_avg_func(avg_type)(sig_magnitude)

    # Identify time periods of bursting using the 2 thresholds
    is_burst = _dual_threshold_split(sig_magnitude, dual_thresh[1], dual_thresh[0])

    # Remove bursts detected that are too short
    min_period_length = int(np.ceil(min_cycles * fs / f_range[0]))
    is_burst = _rmv_short_periods(is_burst, min_period_length)

    return is_burst.astype(bool)


def _dual_threshold_split(sig, thresh_hi, thresh_lo):
    """Identify periods of a time series that are above thresh_lo and
    have at least one value above thresh_hi.
    """

    # Find all values above thresh_hi
    # To avoid bug in later loop, do not allow first or last index to start off as 1
    sig[[0, -1]] = 0
    idx_over_hi = np.where(sig >= thresh_hi)[0]

    # Initialize values in identified period
    positive = np.zeros(len(sig))
    positive[idx_over_hi] = 1

    # Iteratively test if a value is above thresh_lo if it is not currently in an identified period
    sig_len = len(sig)

    for ind in idx_over_hi:

        j_down = ind - 1
        if positive[j_down] == 0:
            j_down_done = False
            while j_down_done is False:
                if sig[j_down] >= thresh_lo:
                    positive[j_down] = 1
                    j_down -= 1
                    if j_down < 0:
                        j_down_done = True
                else:
                    j_down_done = True

        j_up = ind + 1
        if positive[j_up] == 0:
            j_up_done = False
            while j_up_done is False:
                if sig[j_up] >= thresh_lo:
                    positive[j_up] = 1
                    j_up += 1
                    if j_up >= sig_len:
                        j_up_done = True
                else:
                    j_up_done = True

    return positive


def _rmv_short_periods(sig, n_samples):
    """Remove periods that == 1 for less than n_samples."""

    if np.sum(sig) == 0:
        return sig

    osc_changes = np.diff(1 * sig)
    osc_starts = np.where(osc_changes == 1)[0]
    osc_ends = np.where(osc_changes == -1)[0]

    if len(osc_starts) == 0:
        osc_starts = [0]
    if len(osc_ends) == 0:
        osc_ends = [len(osc_changes)]

    if osc_ends[0] < osc_starts[0]:
        osc_starts = np.insert(osc_starts, 0, 0)
    if osc_ends[-1] < osc_starts[-1]:
        osc_ends = np.append(osc_ends, len(osc_changes))

    osc_length = osc_ends - osc_starts
    osc_starts_long = osc_starts[osc_length >= n_samples]
    osc_ends_long = osc_ends[osc_length >= n_samples]

    is_osc = np.zeros(len(sig))
    for ind in range(len(osc_starts_long)):
        is_osc[osc_starts_long[ind]:osc_ends_long[ind]] = 1

    return is_osc



"""Test periodic simulation functions."""

from neurodsp.sim.periodic import *
from neurodsp.sim.periodic import _make_is_osc

###################################################################################################
###################################################################################################

FS = 100
N_SECONDS = 1
FREQ = 10

def test_sim_oscillation():

    sig = sim_oscillation(N_SECONDS, FS, FREQ)
    assert True

def test_sim_bursty_oscillation():

    sig = sim_bursty_oscillation(N_SECONDS, FS, FREQ)
    assert True

## PRIVATE FUNCTIONS

def test_make_is_osc():
    pass



"""Plots for time series, for NeuroDSP."""

from itertools import repeat, cycle

import numpy as np
import numpy.ma as ma
import matplotlib.pyplot as plt

from neurodsp.plts.style import style_plot
from neurodsp.plts.utils import check_ax, savefig

###################################################################################################
###################################################################################################

@savefig
@style_plot
def plot_time_series(times, sigs, labels=None, colors=None, ax=None):
    """Plot a neural time series.

    Parameters
    ----------
    times : 1d array or list of 1d array
        Time definition(s) for the time series to be plotted.
    sigs : 1d array or list of 1d array
        Time series to plot.
    labels : list of str, optional
        Labels for each time series.
    cols : str or list of str
        Colors to use to plot lines.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (15, 3))

    times = repeat(times) if isinstance(times, np.ndarray) else times
    sigs = [sigs] if isinstance(sigs, np.ndarray) else sigs

    if labels is not None:
        labels = [labels] if not isinstance(labels, list) else labels
    else:
        labels = repeat(labels)

    if colors is not None:
        colors = repeat(colors) if not isinstance(colors, list) else cycle(colors)
    else:
        colors = cycle(['k', 'r', 'b', 'g', 'm', 'c'])

    for time, sig, color, label in zip(times, sigs, colors, labels):
        ax.plot(time, sig, color, label=label)

    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Voltage (uV)')


@savefig
@style_plot
def plot_instantaneous_measure(times, sigs, measure='phase', ax=None, **plt_kwargs):
    """Plot an instantaneous measure, of phase, amplitude or frequency.

    Parameters
    ----------
    times : 1d array or list of 1d array
        Time definition(s) for the time series to be plotted.
    sigs : 1d array or list of 1d array
        Time series to plot.
    measure : {'phase', 'amplitude', 'frequency'}
        Which kind of measure is being plotted.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    **kwargs
        Keyword arguments to pass into `plot_time_series`.
    """

    if measure not in ['phase', 'amplitude', 'frequency']:
        raise ValueError('Measure not understood.')

    if measure == 'phase':
        plot_time_series(times, sigs, ax=ax, ylabel='Phase (rad)', **plt_kwargs)
        plt.yticks([-np.pi, 0, np.pi], ['-$\pi$', 0, '$\pi$'])
    elif measure == 'amplitude':
        plot_time_series(times, sigs, ax=ax, ylabel='Amplitude', **plt_kwargs)
    elif measure == 'frequency':
        plot_time_series(times, sigs, ax=ax, ylabel='Instantaneous\nFrequency (Hz)', **plt_kwargs)


@savefig
@style_plot
def plot_bursts(times, sig, bursting, ax=None, **plt_kwargs):
    """Plot a neural time series, labelling detected bursts.

    Parameters
    ----------
    times : 1d array
        Time definition for the time series to be plotted.
    sig : 1d array
        Time series to plot.
    bursting : 1d array
        A boolean array which indicates identified bursts.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    **kwargs
        Keyword arguments to pass into `plot_time_series`.
    """

    ax = check_ax(ax, (15, 3))

    bursts = ma.array(sig, mask=np.invert(bursting))
    plot_time_series(times, [sig, bursts], ax=ax, **plt_kwargs)



"""Tests for FIR filters."""

from pytest import raises

from neurodsp.filt.fir import *

###################################################################################################
###################################################################################################

def test_filter_signal_fir(tsig):

    sig = filter_signal_fir(tsig, 500, 'bandpass', (8, 12))
    assert True

def test_design_fir_filter():

    sig_length, fs = 1000, 100
    test_filts = {'bandpass' : (5, 10), 'bandstop' : (5, 6),
                  'lowpass' : (None, 5), 'highpass' : (5, None)}

    for pass_type, f_range in test_filts.items():
        filter_coefs = design_fir_filter(sig_length, fs, pass_type, f_range)
    assert True

def test_compute_filt_len():

    # Settings for checks
    fs = 500
    sig_length = 2000
    f_lo, f_hi = 4, 8

    # Check filt_len, if defined using n_seconds
    n_seconds = 1.75 # Number chosen to create odd expected filt_len (not needing rounding up)
    expected_filt_len = n_seconds * fs
    filt_len = compute_filt_len(sig_length, fs, 'bandpass', f_lo, f_hi,
                                n_cycles=None, n_seconds=n_seconds)
    assert filt_len == expected_filt_len

    # Check filt_len, if defined using n_cycles
    n_cycles = 5
    expected_filt_len = int(np.ceil(fs * n_cycles / f_lo))
    filt_len = compute_filt_len(sig_length, fs, 'bandpass', f_lo, f_hi,
                                n_cycles=n_cycles, n_seconds=None)
    assert filt_len == expected_filt_len

    # Check filt_len, if expected to be rounded up to be odd
    n_cycles = 4
    expected_filt_len = int(np.ceil(fs * n_cycles / f_lo)) + 1
    filt_len = compute_filt_len(sig_length, fs, 'bandpass', f_lo, f_hi,
                                n_cycles=n_cycles, n_seconds=None)
    assert filt_len == expected_filt_len

    # Check error is raised if the filter designed is longer than the signal
    with raises(ValueError):
        compute_filt_len(1000, fs, 'bandpass', f_lo, f_hi, None, 3)



"""Tests for sim info functions."""

from pytest import raises

from neurodsp.sim.info import *

###################################################################################################
###################################################################################################

def test_get_sim_funcs():

    for module in SIM_MODULES:
        funcs = get_sim_funcs(module)
        assert isinstance(funcs, dict)

def test_get_sim_names():

    for module in SIM_MODULES:
        funcs = get_sim_names(module)
        assert isinstance(funcs, list)

def test_get_sim_func():

    # Check a successful function request
    func = get_sim_func('sim_oscillation')

    # Check the error for requesting non-existing function
    with raises(ValueError):
        get_sim_func('bad_func')



"""Utility functions for burst analysis."""

import numpy as np

###################################################################################################
###################################################################################################

def compute_burst_stats(bursting, fs):
    """Get statistics of bursts.

    Parameters
    ----------
    bursting : 1d array
        Boolean indication of where bursts are present in the input signal.
    fs : float
        Sampling rate, in Hz.

    Returns
    -------
    stats_dict : dict
        Contains the following keys:

        * 'n_bursts': the number of bursts
        * 'duration_mean': mean duration of bursts, in seconds
        * 'duration_std': standard deviation of burst durations, in seconds
        * 'percent_burst': percent time in bursts
        * 'burst_rate': bursts/sec
    """

    tot_time = len(bursting) / fs

    # Find burst starts and ends
    starts = np.array([])
    ends = np.array([])

    for ii, index in enumerate(np.where(np.diff(bursting) != 0)[0]):

        if (ii % 2) == 0:
            starts = np.append(starts, index)
        else:
            ends = np.append(ends, index)

    # Duration of each burst
    durations = (ends - starts) / fs

    stats_dict = {'n_bursts': len(starts),
                  'duration_mean': np.mean(durations),
                  'duration_std': np.std(durations),
                  'percent_burst': 100 * np.sum(bursting) / len(bursting),
                  'bursts_per_second': len(starts) / tot_time}

    return stats_dict



"""Filtering functions."""

from .filter import filter_signal
from .fir import filter_signal_fir, design_fir_filter
from .iir import filter_signal_iir, design_iir_filter
from .utils import infer_passtype



"""Utility functions for neurodsp."""

import numpy as np

###################################################################################################
###################################################################################################

def demean(array, mean=0., select_nonzero=True):
    """Demean an array, updating to specified mean.

    Parameters
    ----------
    array : 1d array
        Data to demean.
    mean : float, optional, default: 0
        New mean for data to have.
    select_nonzero : boolean, optional, default=True
        Whether to calculate the mean of the array across only non-zero data points.

    Returns
    -------
    out : 1d array
        Demeaned data.
    """

    if select_nonzero:
        nonzero = np.nonzero(array)
        nonzero_mean = np.mean(array[nonzero])
        out = array.copy()
        out[nonzero] = array[nonzero] - nonzero_mean + mean

    else:
        out = array - array.mean() + mean

    return out


def normalize_variance(array, variance=1., select_nonzero=True):
    """Normalize the variance of an array, updating to specified variance.

    Parameters
    ----------
    array : 1d array
        Data to normalize variance to.
    variance : float, optional, default=1.
        Variance to normalize to.
    select_nonzero : boolean, optional, default=True
        Whether to calculate the variance of the array across only non-zero data points.

    Returns
    -------
    out : 1d array
        Variance normalized data.

    Notes
    -----
    If the input array is all zeros, this function simply returns the input.
    """

    # If array is all zero, set to return the same array
    #   Can't update variance if it's zero
    if not array.any():
        out = array

    else:

        if select_nonzero:
            nonzero = np.nonzero(array)
            array_std = np.std(array[nonzero])
            out = array.copy()
            out[nonzero] = array[nonzero] / array_std * np.sqrt(variance)
        else:
            out = array / array.std() * np.sqrt(variance)

    return out


def remove_nans(sig):
    """Drop any NaNs on the edges of a 1d array.

    Parameters
    ----------
    sig : 1d array
        Signal to be checked for edge NaNs.

    Returns
    -------
    sig_removed : 1d array
        Signal with NaN edges removed.
    sig_nans : 1d array
        Boolean array indicating where NaNs were in the original array.
    """

    sig_nans = np.isnan(sig)
    sig_removed = sig[np.where(~np.isnan(sig))]

    return sig_removed, sig_nans


def restore_nans(sig, sig_nans, dtype=float):
    """Restore NaN values to the edges of a 1d array.

    Parameters
    ----------
    sig : 1d array
        Signal that has had NaN edges removed.
    sig_nans : 1d array
        Boolean array indicating where NaNs were in the original array.

    Returns
    -------
    sig_restored : 1d array
        Signal with NaN edges restored.
    """

    sig_restored = np.ones(len(sig_nans), dtype=dtype) * np.nan
    sig_restored[~sig_nans] = sig

    return sig_restored


def discard_outliers(data, outlier_percent):
    """Discard outlier arrays with high values.

    Parameters
    ----------
    data : nd array
        Array to remove outliers from.
    outlier_percent : float
        The percentage of outlier values to be removed.

    Returns
    -------
    data : nd array
        Array after removing outliers.
    """

    # Get the number of arrays to discard - round up so it doesn't get a zero.
    n_discard = int(np.ceil(data.shape[-1] / 100. * outlier_percent))

    # Make 2D -> 3D for looping across array
    data = data[np.newaxis, :, :] if data.ndim == 2 else data

    # Select the windows to keep from each 2D component of the input data
    data = [dat[:, np.argsort(np.mean(np.log10(dat), axis=0))[:-n_discard]] for dat in data]

    # Reshape array and squeeze to drop back to 2D if that was original shape
    data = np.squeeze(np.stack(data))

    return data



"""Decorators for neurodsp.sim."""

from functools import wraps

import numpy as np

from neurodsp.utils.utils import demean, normalize_variance

###################################################################################################
###################################################################################################

def normalize(func, **kwargs):
    """Decorator function to normalize the first output of the wrapped function."""

    @wraps(func)
    def decorated(*args, **kwargs):

        # Grab variance & mean as possible kwargs, with default values if not
        variance = kwargs.pop('variance', 1.)
        mean = kwargs.pop('mean', 0.)

        # Call sim function, and unpack to get sig variable, if there are multiple returns
        out = func(*args, **kwargs)
        sig = out[0] if isinstance(out, tuple) else out

        # Apply variance & mean transformations
        if variance is not None:
            sig = normalize_variance(sig, variance=variance)
        if mean is not None:
            sig = demean(sig, mean=mean)

        # Return sig & other outputs, if there were any, or just sig otherwise
        return (sig, out[1:]) if isinstance(out, tuple) else sig

    return decorated


def multidim(func, *args, **kwargs):
    """Decorator function to apply the wrapped function across dimensions."""

    @wraps(func)
    def decorated(sig, *args, **kwargs):

        if sig.ndim == 1:
            out = func(sig, *args, **kwargs)

        elif sig.ndim == 2:
            out = np.vstack([func(dat, *args, **kwargs) for dat in sig])

        return out

    return decorated



"""Test filtering plots."""

from neurodsp.filt.filter import filter_signal
from neurodsp.filt.fir import design_fir_filter
from neurodsp.filt.utils import compute_frequency_response

from neurodsp.tests.utils import plot_test

from neurodsp.plts.filt import *

###################################################################################################
###################################################################################################

@plot_test
def test_plot_filter_properties():

    fs = 1000
    coefs = design_fir_filter(2000, fs, 'bandpass', (8, 12), 3)
    f_db, db = compute_frequency_response(coefs, a_vals=1, fs=1000)

    plot_filter_properties(f_db, db, fs, coefs)

    assert True

@plot_test
def test_plot_frequency_response():

    coefs = design_fir_filter(2000, 1000, 'bandpass', (8, 12), 3)
    f_db, db = compute_frequency_response(coefs, a_vals=1, fs=1000)

    plot_frequency_response(f_db, db)

    assert True

@plot_test
def test_plot_impulse_response():

    fs = 1000
    coefs = design_fir_filter(2000, fs, 'bandpass', (8, 12), 3)

    plot_impulse_response(fs, coefs)

    assert True



"""Style helpers and utilities for NeuroDSP plots."""

from functools import wraps

import matplotlib.pyplot as plt

###################################################################################################
###################################################################################################

PLOT_STYLE_ARGS = ['title', 'xlabel', 'ylabel', 'xlim', 'ylim', 'alpha', 'lw']

def plot_style(ax, **kwargs):
    """Define plot style."""

    # Set any provided plot arguments
    ax.set(**kwargs)

    # Aesthetics and axis labels
    ax.xaxis.label.set_size(16)
    ax.yaxis.label.set_size(16)
    ax.tick_params(axis='both', which='major', labelsize=16)

    # If labels were provided, add a legend
    if ax.get_legend_handles_labels()[0]:
        ax.legend(prop={'size': 12}, loc='best')

    # If a title was provided, update the size
    if ax.get_title():
        ax.title.set_size(20)

    plt.tight_layout()


def style_plot(func, *args, **kwargs):
    """Decorator function to apply a plot style function, after plot generation."""

    @wraps(func)
    def decorated(*args, **kwargs):

        # Grab a custom style function, if provided, and grab any provided style arguments
        style_func = kwargs.pop('plot_style', plot_style)
        style_kwargs = {key : kwargs.pop(key) for key in PLOT_STYLE_ARGS if key in kwargs}

        # Create the plot
        func(*args, **kwargs)

        # Get plot axis, if a specific one was provided, or just grab current and apply style
        cur_ax = kwargs['ax'] if 'ax' in kwargs and kwargs['ax'] is not None else plt.gca()
        style_func(cur_ax, **style_kwargs)

    return decorated



"""Tests for utility functions."""

import numpy as np
from numpy.testing import assert_equal

from neurodsp.utils.utils import *

###################################################################################################
###################################################################################################

def test_demean():

    d1 = np.array([1, 2, 3])
    d2 = np.array([0, 1, 2, 3, 0])

    # Check default - demean to 0
    out1 = demean(d1)
    assert np.isclose(out1.mean(), 0.)

    # Check demeaning and adding specific mean
    out2 = demean(d1, mean=1.)
    assert np.isclose(out2.mean(), 1.)

    # Check dealing with zero entries
    out3 = demean(d2)
    assert np.isclose(out3[np.nonzero(out3)].mean(), 0)

    # Check turning of non-zero selection
    out3 = demean(d2, mean=1, select_nonzero = False)
    assert np.isclose(out3.mean(), 1)

def test_normalize_variance():

    d1 = np.array([1, 2, 3])
    d2 = np.array([0, 1, 2, 3, 0])

    # Check default - normalize variance to 1
    out1 = normalize_variance(d1)
    np.isclose(out1.var(), 1.)

    # Check normalizing and add specific variance
    out2 = normalize_variance(d1, 2.)
    np.isclose(out2.var(), 2.)

def test_remove_nans():

    # Test with equal # of NaNs on either edge
    arr = np.array([np.NaN, np.NaN, 1, 2, 3, np.NaN, np.NaN])
    arr_no_nans, arr_nans = remove_nans(arr)
    assert_equal(arr_no_nans, np.array([1, 2, 3]))
    assert_equal(arr_nans, np.array([True, True, False, False, False, True, True]))

    # Test with different # of NaNs on either edge
    arr = np.array([np.NaN, np.NaN, 1, 2, 3, 4, np.NaN,])
    arr_no_nans, arr_nans = remove_nans(arr)
    assert_equal(arr_no_nans, np.array([1, 2, 3, 4]))
    assert_equal(arr_nans, np.array([True, True, False, False, False, False, True]))

def test_restore_nans():

    arr_no_nans = np.array([1, 2, 3])
    arr_nans =  np.array([True, True, False, False, False, True])

    arr_restored = restore_nans(arr_no_nans, arr_nans)
    assert_equal(arr_restored, np.array([np.NaN, np.NaN, 1, 2, 3, np.NaN]))

def test_discard_outliers():
    pass



"""Test functions in the spectral domain analysis module."""

from neurodsp.spectral.spectral import *

###################################################################################################
###################################################################################################

def test_compute_spectrum(tsig):

    freqs, spectrum = compute_spectrum(tsig, fs=500, method='welch')
    freqs, spectrum = compute_spectrum(tsig, fs=500, method='wavelet', freqs=[5, 10, 15])
    freqs, spectrum = compute_spectrum(tsig, fs=500, method='medfilt')
    assert True

def test_compute_spectrum_welch(tsig):

    freqs, spectrum = compute_spectrum_welch(tsig, fs=500, avg_type='mean')
    freqs, spectrum = compute_spectrum_welch(tsig, fs=500, avg_type='median')
    assert True

def test_compute_spectrum_wavelet(tsig):

    freqs, spectrum = compute_spectrum_wavelet(tsig, fs=500, freqs=[5, 10, 15], avg_type='mean')
    freqs, spectrum = compute_spectrum_wavelet(tsig, fs=500, freqs=[5, 10, 15], avg_type='median')
    assert True

def test_compute_spectrum_medfilt(tsig):

    # NOTE: this test fails uninformatively if fs is 1000 (for siglength 1000). Something to check

    freqs, spectrum = compute_spectrum_medfilt(tsig, fs=500)
    assert True

def test_compute_scv(tsig):

    freqs, spect_cv = compute_scv(tsig, fs=500)
    assert True

def test_compute_scv_rs(tsig):

    freqs, t_inds, spect_cv = compute_scv_rs(tsig, fs=500, method='bootstrap')
    freqs, t_inds, spect_cv = compute_scv_rs(tsig, fs=500, method='rolling')
    assert True

def test_compute_spectral_hist(tsig):

    freqs, bins, spectral_hist = compute_spectral_hist(tsig, fs=500)
    assert True

## PRIVATE FUNCTIONS
def test_spg_settings():
    pass



"""Utility functions for filtering."""

import numpy as np
from scipy.signal import freqz

from neurodsp.filt.checks import check_filter_definition

###################################################################################################
###################################################################################################

def infer_passtype(f_range):
    """Given frequency definition of a filter, infer the passtype.

    Parameters
    ----------
    f_range : tuple of (float, float)
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.

    Returns
    -------
    pass_type : str
        Which kind of filter pass_type is consistent with the frequency definition provided.

    Notes
    -----
    Assumes that a definition with two frequencies is a 'bandpass' (not 'bandstop').
    """

    if f_range[0] is None:
        pass_type = 'lowpass'
    elif f_range[1] is None:
        pass_type = 'highpass'
    else:
        pass_type = 'bandpass'

    # Check the inferred passtype & frequency definition is valid
    _ = check_filter_definition(pass_type, f_range)

    return pass_type


def compute_frequency_response(b_vals, a_vals, fs):
    """Compute the frequency response of a filter.

    Parameters
    ----------
    b_vals : 1d array
        B value filter coefficients for a filter.
    a_vals : 1d array
        A value filter coefficients for a filter.
    fs : float
        Sampling rate, in Hz.

    Returns
    -------
    f_db : 1d array
        Frequency vector corresponding to attenuation decibels, in Hz.
    db : 1d array
        Degree of attenuation for each frequency specified in f_db, in dB.
    """

    w_vals, h_vals = freqz(b_vals, a_vals)
    f_db = w_vals * fs / (2. * np.pi)
    db = 20 * np.log10(abs(h_vals))

    return f_db, db


def compute_pass_band(fs, pass_type, f_range):
    """Compute the pass bandwidth of a filter.

    Parameters
    ----------
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.

    Returns
    -------
    pass_bw : float
        The pass bandwidth of the filter.
    """

    f_lo, f_hi = check_filter_definition(pass_type, f_range)
    if pass_type in ['bandpass', 'bandstop']:
        pass_bw = f_hi - f_lo
    elif pass_type == 'highpass':
        pass_bw = compute_nyquist(fs) - f_lo
    elif pass_type == 'lowpass':
        pass_bw = f_hi

    return pass_bw


def compute_transition_band(f_db, db, low=-20, high=-3):
    """Compute transition bandwidth of a filter.

    Parameters
    ----------
    f_db : 1d array
        Frequency vector corresponding to attenuation decibels, in Hz.
    db : 1d array
        Degree of attenuation for each frequency specified in f_db, in dB.
    low : float, optional, default: -20
        The lower limit that defines the transition band, in dB.
    high : float, optional, default: -3
        The upper limit that defines the transition band, in dB.

    Returns
    -------
    transition_band : float
        The transition bandwidth of the filter.
    """

    # This gets the indices of transitions to the values in searched for range
    inds = np.where(np.diff(np.logical_and(db > low, db < high)))[0]
    # This steps through the indices, in pairs, selecting from the vector to select from
    transition_band = np.max([(b - a) for a, b in zip(f_db[inds[0::2]], f_db[inds[1::2]])])

    return transition_band


def compute_nyquist(fs):
    """Compute the nyquist frequency.

    Parameters
    ----------
    fs : float
        Sampling rate, in Hz.

    Returns
    -------
    float
        The nyquist frequency of a signal with the given sampling rate, in Hz.
    """

    return fs / 2.


def remove_filter_edges(sig, filt_len):
    """Drop the edges, by making NaN, from a filtered signal, to avoid edge artifacts.

    Parameters
    ----------
    sig : 1d array
        Filtered signal to have edge artifacts removed from.
    filt_len : int
        Length of the filter that was applied.

    Returns
    -------
    sig : 1d array
        Filter signal with edge artifacts switched to NaNs.
    """

    n_rmv = int(np.ceil(filt_len / 2))
    sig[:n_rmv] = np.nan
    sig[-n_rmv:] = np.nan

    return sig



"""Plotting functions for neurodsp.filt."""

import numpy as np
import matplotlib.pyplot as plt

from neurodsp.plts.style import style_plot
from neurodsp.plts.utils import check_ax, savefig

###################################################################################################
###################################################################################################

@savefig
def plot_filter_properties(f_db, db, fs, impulse_response):
    """Plot filter properties, including frequency response and filter kernel.

    Parameters
    ----------
    f_db : 1d array
        Frequency vector corresponding to attenuation decibels, in Hz.
    db : 1d array
        Degree of attenuation for each frequency specified in f_db, in dB.
    impulse_response : 1d array
        The impulse response of a filter. For an FIR filter, these are the filter coefficients.
    """

    fig, ax = plt.subplots(1, 2, figsize=(10, 5))

    plot_frequency_response(f_db, db, ax=ax[0])
    plot_impulse_response(fs, impulse_response, ax=ax[1])


@savefig
@style_plot
def plot_frequency_response(f_db, db, ax=None):
    """Plot the frequency response of a filter.

    Parameters
    ----------
    f_db : 1d array
        Frequency vector corresponding to attenuation decibels, in Hz.
    db : 1d array
        Degree of attenuation for each frequency specified in f_db, in dB.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (5, 5))

    ax.plot(f_db, db, 'k')

    ax.set_title('Frequency response')
    ax.set_xlabel('Frequency (Hz)')
    ax.set_ylabel('Attenuation (dB)')


@savefig
@style_plot
def plot_impulse_response(fs, impulse_response, ax=None):
    """Plot the impulse response of a filter.

    Parameters
    ----------
    impulse_response : 1d array
        The impulse response of a filter.
    ax : matplotlib.Axes, optional
        Figure axes upon which to plot.
    """

    ax = check_ax(ax, (5, 5))

    # Create a samples vector, center to zero, and convert to time
    samples = np.arange(len(impulse_response))
    samples = samples - (len(samples) - 1) / 2
    time = samples / fs

    ax.plot(time, impulse_response, 'k')

    ax.set_title('Kernel')
    ax.set_xlabel('Time (seconds)')
    ax.set_ylabel('Response')



"""Utility functions for NeuroDSP plots."""

from functools import wraps
from os.path import join as pjoin

import matplotlib.pyplot as plt

###################################################################################################
###################################################################################################

def check_ax(ax, figsize=None):
    """Check whether a figure axes object is defined, define if not.

    Parameters
    ----------
    ax : matplotlib.Axes or None
        Axes object to check if is defined.

    Returns
    -------
    ax : matplotlib.Axes
        Figure axes object to use.
    """

    if not ax:
        _, ax = plt.subplots(figsize=figsize)

    return ax


def savefig(func):

    @wraps(func)
    def decorated(*args, **kwargs):

        save_fig = kwargs.pop('save_fig', False)
        file_name = kwargs.pop('file_name', None)
        file_path = kwargs.pop('file_path', None)

        func(*args, **kwargs)

        if save_fig:
            full_path = pjoin(file_path, file_name) if file_path else file_name
            plt.savefig(full_path)

    return decorated



"""Tests for filter check functions."""

from pytest import raises

from neurodsp.filt.fir import design_fir_filter
from neurodsp.filt.checks import *

###################################################################################################
###################################################################################################

def test_check_filter_definition():

    # Check that error catching works for bad pass_type definition
    with raises(ValueError):
        check_filter_definition('not_a_filter', (12, 12))

    # Check that filter definitions that are legal evaluate as expected
    #   Float or partially filled tuple for f_range should work passable
    f_lo, f_hi = check_filter_definition('bandpass', f_range=(8, 12))
    assert f_lo == 8; assert f_hi == 12
    f_lo, f_hi = check_filter_definition('bandstop', f_range=(58, 62))
    assert f_lo == 58; assert f_hi == 62
    f_lo, f_hi = check_filter_definition('lowpass', f_range=58)
    assert f_lo == None ; assert f_hi == 58
    f_lo, f_hi = check_filter_definition('lowpass', f_range=(0, 58))
    assert f_lo == None ; assert f_hi == 58
    f_lo, f_hi = check_filter_definition('highpass', f_range=58)
    assert f_lo == 58 ; assert f_hi == None
    f_lo, f_hi = check_filter_definition('highpass', f_range=(58, 100))
    assert f_lo == 58 ; assert f_hi == None

    # Check that a bandpass & bandstop definitions fail without proper definitions
    with raises(ValueError):
        f_lo, f_hi = check_filter_definition('bandpass', f_range=8)
    with raises(ValueError):
        f_lo, f_hi = check_filter_definition('bandstop', f_range=58)

    # Check that frequencies cannot be inverted
    with raises(ValueError):
        f_lo, f_hi = check_filter_definition('bandpass', f_range=(8, 4))
    with raises(ValueError):
        f_lo, f_hi = check_filter_definition('bandstop', f_range=(62, 58))

def test_check_filter_properties():

    filter_coefs = design_fir_filter(1000, 500, 'bandpass', (8, 12))
    check_filter_properties(filter_coefs, 1, 500, 'bandpass', (8, 12))
    assert True



"""Core (internal) utility functions for NeuroDSP."""

import numpy as np

###################################################################################################
###################################################################################################

def get_avg_func(avg_type):
    """Select a function to use for averaging.

    Parameters
    ----------
    avg_type : {'mean', 'median'}
        The type of averaging function to use.

    Returns
    -------
    avg_func : callable
        Requested averaging function.
    """

    if avg_type == 'mean':
        avg_func = np.mean
    elif avg_type == 'median':
        avg_func = np.median
    else:
        raise ValueError('Averaging method not understood.')

    return avg_func



"""Checker functions for filtering."""

from warnings import warn

import numpy as np

###################################################################################################
###################################################################################################

def check_filter_definition(pass_type, f_range):
    """Check a filter definition for validity, and get f_lo and f_hi.

    Parameters
    ----------
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.

    Returns
    -------
    f_lo : float or None
        The lower frequency range of the filter, specifying the highpass frequency, if specified.
    f_hi : float or None
        The higher frequency range of the filter, specifying the lowpass frequency, if specified.
    """

    if pass_type not in ['bandpass', 'bandstop', 'lowpass', 'highpass']:
        raise ValueError('Filter passtype not understood.')

    ## Check that frequency cutoff inputs are appropriate
    # For band filters, 2 inputs required & second entry must be > first
    if pass_type in ('bandpass', 'bandstop'):
        if isinstance(f_range, tuple) and f_range[0] >= f_range[1]:
            raise ValueError('Second cutoff frequency must be greater than first.')
        elif isinstance(f_range, (int, float)) or len(f_range) != 2:
            raise ValueError('Two cutoff frequencies required for bandpass and bandstop filters')

        # Map f_range to f_lo and f_hi
        f_lo, f_hi = f_range

    # For lowpass and highpass can be tuple or int/float
    if pass_type == 'lowpass':
        if isinstance(f_range, (int, float)):
            f_hi = f_range
        elif isinstance(f_range, tuple):
            f_hi = f_range[1]
        f_lo = None

    if pass_type == 'highpass':
        if isinstance(f_range, (int, float)):
            f_lo = f_range
        elif isinstance(f_range, tuple):
            f_lo = f_range[0]
        f_hi = None

    # Make sure pass freqs are floats
    f_lo = float(f_lo) if f_lo else f_lo
    f_hi = float(f_hi) if f_hi else f_hi

    return f_lo, f_hi


def check_filter_properties(b_vals, a_vals, fs, pass_type, f_range, transitions=(-20, -3), verbose=True):
    """Check a filters properties, including pass band and transition band.

    Parameters
    ----------
    b_vals : 1d array
        B value filter coefficients for a filter.
    a_vals : 1d array
        A value filter coefficients for a filter.
    fs : float
        Sampling rate, in Hz.
    pass_type : {'bandpass', 'bandstop', 'lowpass', 'highpass'}
        Which kind of filter to apply:

        * 'bandpass': apply a bandpass filter
        * 'bandstop': apply a bandstop (notch) filter
        * 'lowpass': apply a lowpass filter
        * 'highpass' : apply a highpass filter
    f_range : tuple of (float, float) or float
        Cutoff frequency(ies) used for filter, specified as f_lo & f_hi.
        For 'bandpass' & 'bandstop', must be a tuple.
        For 'lowpass' or 'highpass', can be a float that specifies pass frequency, or can be
        a tuple and is assumed to be (None, f_hi) for 'lowpass', and (f_lo, None) for 'highpass'.
    transitions : tuple of (float, float), optional, default: (-20, -3)
        Cutoffs, in dB, that define the transition band.
    verbose : bool, optional, default: True
        Whether to print out transition and pass bands.

    Returns
    -------
    passes : bool
        Whether all the checks pass. False if one or more checks fail.
    """

    # Import utility functions inside function to avoid circular imports
    from neurodsp.filt.utils import (compute_frequency_response,
                                     compute_pass_band, compute_transition_band)

    # Initialize variable to keep track if all checks pass
    passes = True

    # Compute the frequency response
    f_db, db = compute_frequency_response(b_vals, a_vals, fs)

    # Check that frequency response goes below transition level (has significant attenuation)
    if np.min(db) >= transitions[0]:
        passes = False
        warn('The filter attenuation never goes below {} dB.'\
             'Increase filter length.'.format(transitions[0]))
        # If there is no attenuation, cannot calculate bands, so return here
        return passes

    # Check that both sides of a bandpass have significant attenuation
    if pass_type == 'bandpass':
        if db[0] >= transitions[0] or db[-1] >= transitions[0]:
            passes = False
            warn('The low or high frequency stopband never gets attenuated by'\
                 'more than {} dB. Increase filter length.'.format(abs(transitions[0])))

    # Compute pass & transition bandwidth
    pass_bw = compute_pass_band(fs, pass_type, f_range)
    transition_bw = compute_transition_band(f_db, db, transitions[0], transitions[1])

    # Raise warning if transition bandwidth is too high
    if transition_bw > pass_bw:
        passes = False
        warn('Transition bandwidth is  {:.1f}  Hz. This is greater than the desired'\
             'pass/stop bandwidth of  {:.1f} Hz'.format(transition_bw, pass_bw))

    # Print out transition bandwidth and pass bandwidth to the user
    if verbose:
        print('Transition bandwidth is {:.1f} Hz.'.format(transition_bw))
        print('Pass/stop bandwidth is {:.1f} Hz.'.format(pass_bw))

    return passes



"""Tools for estimating time frequency properties, using hilbert methods."""

import numpy as np
from scipy.signal import hilbert

from neurodsp.filt import filter_signal
from neurodsp.utils.decorators import multidim
from neurodsp.utils.utils import remove_nans, restore_nans
from neurodsp.filt.utils import infer_passtype, remove_filter_edges

###################################################################################################
###################################################################################################

@multidim
def robust_hilbert(sig, increase_n=False):
    """Compute the hilbert transform, ignoring the boundaries of that are filled with NaN.

    Parameters
    ----------
    sig : 1d array
        Time series.
    increase_n : bool, optional, default: False
        If True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of x.

    Returns
    -------
    sig_hilb : 1d array
        The hilbert transform of the input signal.
    """

    # Extract the signal that is not nan
    sig_nonan, sig_nans = remove_nans(sig)

    # Compute hilbert transform of signal without nans
    if increase_n:
        sig_len = len(sig_nonan)
        n_components = 2**(int(np.log2(sig_len)) + 1)
        sig_hilb_nonan = hilbert(sig_nonan, n_components)[:sig_len]
    else:
        sig_hilb_nonan = hilbert(sig_nonan)

    # Fill in output hilbert with nans on edges
    sig_hilb = restore_nans(sig_hilb_nonan, sig_nans, dtype=complex)

    return sig_hilb


@multidim
def phase_by_time(sig, fs, f_range, hilbert_increase_n=False, remove_edges=True, **filter_kwargs):
    """Calculate the phase time series of a neural oscillation.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of float
        Frequency range, in Hz, as (low, high).
    hilbert_increase_n : bool, optional, default: False
        If True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of x.
    remove_edges : bool, optional, default: True
        If True, replace the samples that are within half a kernel's length to
        the signal edge with np.nan.
    **filter_kwargs
        Keyword parameters to pass to `filter_signal`.

    Returns
    -------
    pha : 1d array
        Time series of phase.
    """

    sig_filt, kernel = filter_signal(sig, fs, infer_passtype(f_range), f_range=f_range,
                                     remove_edges=False, return_filter=True, **filter_kwargs)

    pha = np.angle(robust_hilbert(sig_filt, increase_n=hilbert_increase_n))

    if remove_edges:
        pha = remove_filter_edges(pha, len(kernel))

    return pha


@multidim
def amp_by_time(sig, fs, f_range, hilbert_increase_n=False, remove_edges=True, **filter_kwargs):
    """Calculate the amplitude time series.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of float
        The frequency filtering range, in Hz, as (low, high).
    hilbert_increase_n : bool, optional, default: False
        If True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of sig.
    remove_edges : bool, optional, default: True
        If True, replace the samples that are within half a kernel's length to
        the signal edge with np.nan.
    **filter_kwargs
        Keyword parameters to pass to `filter_signal`.

    Returns
    -------
    amp : 1d array
        Time series of amplitude.
    """

    sig_filt, kernel = filter_signal(sig, fs, infer_passtype(f_range), f_range=f_range,
                                     remove_edges=False, return_filter=True, **filter_kwargs)

    amp = np.abs(robust_hilbert(sig_filt, increase_n=hilbert_increase_n))

    if remove_edges:
        amp = remove_filter_edges(amp, len(kernel))

    return amp


@multidim
def freq_by_time(sig, fs, f_range, hilbert_increase_n=False, remove_edges=True, **filter_kwargs):
    """Estimate the instantaneous frequency at each sample.

    Parameters
    ----------
    sig : 1d array
        Voltage time series.
    fs : float
        Sampling rate, in Hz.
    f_range : tuple of float
        Frequency range for filtering, in Hz, as (low, high).
    hilbert_increase_n : bool, optional, default: False
        If True, zeropad the signal to length the next power of 2 when doing the hilbert transform.
        This is because scipy.signal.hilbert can be very slow for some lengths of sig.
    remove_edges : bool, optional, default: True
        If True, replace the samples that are within half a kernel's length to
        the signal edge with np.nan.
    **filter_kwargs
        Keyword parameters to pass to `filter_signal`.

    Returns
    -------
    i_f : float
        Estimated instantaneous frequency for each sample in 'sig'.

    Notes
    -----
    This function assumes monotonic phase, so a phase slip will
    be processed as a very high frequency.
    """

    pha = phase_by_time(sig, fs, f_range, hilbert_increase_n,
                        remove_edges, **filter_kwargs)

    phadiff = np.diff(pha)
    phadiff[phadiff < 0] = phadiff[phadiff < 0] + 2 * np.pi

    i_f = fs * phadiff / (2 * np.pi)
    i_f = np.insert(i_f, 0, np.nan)

    return i_f



"""Simulating time series, with combinations of periodic, aperiodic and transient components."""

from itertools import repeat

import numpy as np

from neurodsp.utils.decorators import normalize
from neurodsp.sim.info import get_sim_func

###################################################################################################
###################################################################################################

@normalize
def sim_combined(n_seconds, fs, simulations, variances=1):
    """Sim multiple component signals and combine them.

    Parameters
    ----------
    n_seconds : float
        Signal duration, in seconds.
    fs : float
        Signal sampling rate, in Hz.
    simulations : dictionary
        A dictionary of simulation functions to run, with their desired parameters.
    variances : list of float or 1
        Specified variance for each component of the signal.

    Returns
    -------
    sig : 1d array
        Simulated combined signal.
    """

    if not (variances == 1 or len(variances) == len(simulations)):
        raise ValueError('Simulations and proportions lengths do not match.')

    # Collect the sim function to use, and repeat variance if is set to 1
    simulations = {(get_sim_func(name) if isinstance(name, str) else name) : params \
                   for name, params in simulations.items()}
    variances = repeat(variances) if isinstance(variances, int) else variances

    # Simulate each component, specifying variance, and combine them
    components = [func(n_seconds, fs, **params, variance=variance) for \
        (func, params), variance in zip(simulations.items(), variances)]
    sig = np.sum(components, axis=0)

    return sig



"""Test time series plots."""

from pytest import raises

from neurodsp.tests.utils import plot_test

from neurodsp.plts.time_series import *

###################################################################################################
###################################################################################################

@plot_test
def test_plot_time_series(tsig):

    times = np.arange(0, len(tsig), 1)
    plot_time_series(times, tsig)

@plot_test
def test_plot_instantaneous_measure(tsig):

    times = np.arange(0, len(tsig), 1)
    plot_instantaneous_measure(times, tsig, 'phase')
    plot_instantaneous_measure(times, tsig, 'amplitude')
    plot_instantaneous_measure(times, tsig, 'frequency')

    # Check the error for bad measure
    with raises(ValueError):
        plot_instantaneous_measure(times, tsig, 'BAD')

@plot_test
def test_plot_bursts(tsig):

    times = np.arange(0, len(tsig), 1)
    bursts = np.array([True] * len(tsig))
    plot_bursts(times, tsig, bursts)



"""Test aperiodic simulation functions."""

from neurodsp.sim.aperiodic import *

###################################################################################################
###################################################################################################

FS = 100
N_SECONDS = 1

def test_sim_poisson_pop():

    sig = sim_poisson_pop(N_SECONDS, FS)
    assert True

def test_sim_synaptic_current():

    sig = sim_synaptic_current(N_SECONDS, FS)
    assert True

def test_sim_random_walk():

    sig = sim_random_walk(N_SECONDS, FS)
    assert True

def test_powerlaw():

    sig = sim_powerlaw(N_SECONDS, FS)
    assert True



"""NeuroDSP setup script."""

import os
from setuptools import setup, find_packages

# Get the current version number from inside the module
with open(os.path.join('neurodsp', 'version.py')) as vf:
    exec(vf.read())

# Copy in long description.
#  Note: this is a partial copy from the README
#    Only update here in coordination with the README, to keep things consistent.
long_description = \
"""
========
Neurodsp
========

A package for digital signal processing of neural time series.

Neurodsp contains several modules:

- burst : Detect bursting oscillators in neural signals
- filt : Filter data with bandpass, highpass, lowpass, or notch filters
- laggedcoherence : Estimate rhythmicity using the lagged coherence measure
- sim : Simulate bursting or stationary oscillators with brown noise
- spectral : Compute spectral domain features (PSD and 1/f slope, etc)
- swm : Identify recurrent patterns in a signal using sliding window matching
- timefrequency : Estimate instantaneous measures of oscillatory activity
"""

setup(
    name = 'neurodsp',
    version = __version__,
    description = 'Digital Signal Processing for Neural time series',
    long_description = long_description,
    author = 'The Voytek Lab',
    author_email = 'voyteklab@gmail.com',
    url = 'https://github.com/neurodsp-tools/neurodsp',
    packages = find_packages(),
    license = 'MIT',
    download_url = 'https://github.com/neurodsp-tools/neurodsp/releases',
    keywords = ['neuroscience', 'neural oscillations', 'time series analysis', 'spectral analysis', 'LFP'],
    install_requires = ['numpy', 'scipy', 'matplotlib', 'pandas', 'scikit-learn'],
    tests_require = ['pytest'],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
        ]
)



"""Spectral module, for calculating power spectra and associated functionality."""

from .spectral import (compute_spectrum, compute_spectrum_welch,
                       compute_spectrum_wavelet, compute_spectrum_medfilt)
from .spectral import compute_scv, compute_scv_rs, compute_spectral_hist
from .utils import trim_spectrum, rotate_powerlaw



"""Simulating transients."""

from warnings import warn

import numpy as np
from scipy.signal import gaussian, sawtooth

###################################################################################################
###################################################################################################

def sim_osc_cycle(n_seconds, fs, cycle_type, **cycle_params):
    """Make one cycle of an oscillation.

    Parameters
    ----------
    n_seconds : float
        Length of cycle window in seconds.
        Note that this is NOT the period of the cycle, but the length of the returned array
        that contains the cycle, which can be (and usually is) much shorter.
    fs : float
        Sampling frequency of the cycle simulation.
    cycle_type : {'sine', 'asine', 'sawtooth', 'gaussian', 'exp', '2exp'}
        What type of cycle to simulate. Options:

        * sine: a sine wave cycle
        * asine: an asymmetric sine wave
        * sawtooth: a sawtooth wave
        * gaussian: a gaussian cycle
        * exp: a cycle with exponential decay
        * 2exp: a cycle with exponential rise and decay

    **cycle_params
        Keyword arguments for parameters of the oscillation cycle, all as float:

        * sine: None
        * asine: 'rdsym', rise-decay symmetry, from 0-1
        * sawtooth: 'width', width of the rising ramp as a proportion of the total cycle
        * gaussian: 'std', standard deviation of the gaussian kernel, in seconds
        * exp: 'tau_d', decay time, in seconds
        * 2exp: 'tau_r' & 'tau_d' rise time, and decay time, in seconds

    Returns
    -------
    cycle: 1d array
        Simulated oscillation cycle.
    """

    if cycle_type not in ['sine', 'asine', 'sawtooth', 'gaussian', 'exp', '2exp']:
        raise ValueError('Did not recognize cycle type.')

    if cycle_type == 'sine':
        cycle = np.sin(create_cycle_time(n_seconds, fs))

    elif cycle_type == 'asine':
        cycle = sim_asine_cycle(n_seconds, fs, cycle_params['rdsym'])

    elif cycle_type == 'sawtooth':
        cycle = sawtooth(create_cycle_time(n_seconds, fs), cycle_params['width'])

    elif cycle_type == 'gaussian':
        cycle = gaussian(n_seconds * fs, cycle_params['std'] * fs)

    elif cycle_type == 'exp':
        cycle = sim_synaptic_kernel(n_seconds, fs, 0, cycle_params['tau_d'])

    elif cycle_type == '2exp':
        cycle = sim_synaptic_kernel(n_seconds, fs, cycle_params['tau_r'], cycle_params['tau_d'])

    return cycle


def sim_asine_cycle(n_seconds, fs, rdsym):
    """Simulate a cycle of an asymmetric sine wave.

    Parameters
    ----------
    n_seconds : float
        Length of cycle window in seconds.
        Note that this is NOT the period of the cycle, but the length of the returned array
        that contains the cycle, which can be (and usually is) much shorter.
    fs : float
        Sampling frequency of the cycle simulation.
    rdsym : float
        Rise-decay symmetry of the oscillation, as fraction of the period in the rise time, where:
        = 0.5 - symmetric (sine wave)
        < 0.5 - shorter rise, longer decay
        > 0.5 - longer rise, shorter decay

    Returns
    -------
    cycle : 1d array
        Simulated oscillation cycle.
    """

    # Determine number of samples in rise and decay periods
    n_samples = int(n_seconds * fs)
    n_rise = int(np.round(n_samples * rdsym))
    n_decay = n_samples - n_rise

    # Make phase array for the cycle, and convert to signal
    #   Note: the ceil & floor are so the cycle has the right number of samples if n_decay is odd
    cycle = np.sin(np.hstack([np.linspace(0, np.pi/2, int(np.ceil(n_rise/2)) + 1),
                              np.linspace(np.pi/2, -np.pi/2, n_decay + 1)[1:-1],
                              np.linspace(-np.pi/2, 0, int(np.floor(n_rise/2)) + 1)[:-1]]))

    return cycle


def sim_synaptic_kernel(n_seconds, fs, tau_r, tau_d):
    """Creates synaptic kernels that with specified time constants.

    Parameters
    ----------
    n_seconds : float
        Length of simulated kernel in seconds.
    fs : float
        Sampling rate of simulated signal, in Hz
    tau_r : float
        Rise time of synaptic kernel, in seconds.
    tau_d : float
        Decay time of synaptic kernel, in seconds.

    Returns
    -------
    kernel : 1d array
        Computed synaptic kernel with length equal to t

    Notes
    -----
    3 types of kernels are available, based on combinations of time constants:
    - tau_r == tau_d  : alpha (function) synapse
    - tau_r = 0       : instantaneous rise, (single) exponential decay
    - tau_r!=tau_d!=0 : double-exponential (rise and decay)
    """

    # NOTE: sometimes n_seconds is not exact, resulting in a slightly longer or
    #   shorter times vector, which will affect final signal length
    #   https://docs.python.org/2/tutorial/floatingpoint.html
    times = np.arange(0, n_seconds, 1./fs)

    # Kernel type: single exponential
    if tau_r == 0:
        kernel = np.exp(-times / tau_d)

    # Kernel type: alpha
    elif tau_r == tau_d:
        # I(t) = t/tau * exp(-t/tau)
        kernel = (times / tau_r) * np.exp(-times / tau_r)

    # Kernel type: double exponential
    else:
        if tau_r > tau_d:
            warn('Rise time constant should be shorter than decay time constant.')

        # I(t)=(tau_r/(tau_r-tau_d))*(exp(-t/tau_d)-exp(-t/tau_r))
        kernel = (np.exp(-times / tau_d) - np.exp(-times / tau_r))

    # Normalize the integral to 1
    kernel = kernel / np.sum(kernel)

    return kernel


def create_cycle_time(n_seconds, fs):
    """Create a vector of time indices for a single cycle.

    Parameters
    ----------
    n_seconds : float
        Length of simulated kernel in seconds.
    fs : float
        Sampling rate of simulated signal, in Hz

    Returns
    -------
    1d array
        Time indices.
    """

    return 2*np.pi*1/n_seconds * (np.arange(fs*n_seconds)/fs)



"""Tests for burst detection functions."""

from neurodsp.burst.dualthresh import *
from neurodsp.burst.dualthresh import _dual_threshold_split, _rmv_short_periods

###################################################################################################
###################################################################################################

def test_detect_bursts_dual_threshold(tsig):

    # Test default settings
    bursts = detect_bursts_dual_threshold(tsig, 500, (8, 12), (1, 2))
    # Test other settings
    bursts = detect_bursts_dual_threshold(tsig, 500, (8, 12), (1, 2),
                                          avg_type='mean', magnitude_type='power')
    assert True

## PRIVATE FUNCTION TESTS
def test_dual_threshold_split():
    # TODO
    pass

def tests_rmv_short_periods():
    pass
    #TODO: check this function, doesn't seem to work quite as expected.

    # dat = np.array([0, 1, 0, 1, 1, 0, 1, 1, 1, 0])

    # dat = _rmv_short_periods(dat, 3)
    # assert np.array_equal(dat, np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 0]))



"""Test transient simulation functions."""

from pytest import raises

from neurodsp.sim.transients import *

###################################################################################################
###################################################################################################

N_SECONDS = 1
FS = 100

def test_sim_osc_cycle():

    cycle = sim_osc_cycle(N_SECONDS, FS, 'sine')
    cycle = sim_osc_cycle(N_SECONDS, FS, 'asine', rdsym=0.75)
    cycle = sim_osc_cycle(N_SECONDS, FS, 'sawtooth', width=0.5)
    cycle = sim_osc_cycle(N_SECONDS, FS, 'gaussian', std=2)
    cycle = sim_osc_cycle(N_SECONDS, FS, 'exp', tau_d=0.2)
    cycle = sim_osc_cycle(N_SECONDS, FS, '2exp', tau_r=0.2, tau_d=0.2)

    with raises(ValueError):
        sim_osc_cycle(N_SECONDS, FS, 'not_a_cycle')

def test_asine_cycle():

    cycle = sim_asine_cycle(N_SECONDS, FS, 0.25)
    assert True

def test_sim_make_synaptic_kernel():

    np.random.seed(0)

    # smoke test that valid parameter configurations do not return negative values
    t_ker, tau_r, tau_d = 1., 0., 0.02
    assert np.all(sim_synaptic_kernel(t_ker, FS, tau_r, tau_d) >= 0.)

    t_ker, tau_r, tau_d = 2., 0.005, 0.02
    assert np.all(sim_synaptic_kernel(t_ker, FS, tau_r, tau_d) >= 0.)

    t_ker, tau_r, tau_d = 1., 0.02, 0.02
    assert np.all(sim_synaptic_kernel(t_ker, FS, tau_r, tau_d) >= 0.)

def test_create_cycle_time():

    times = create_cycle_time(N_SECONDS, FS)
    assert True



"""The lagged coherence algorithm for estimating the rhythmicity of a neural signal."""

import numpy as np

from scipy.signal.windows import hann
from neurodsp.utils.decorators import multidim

###################################################################################################
###################################################################################################

@multidim
def lagged_coherence(sig, f_range, fs, n_cycles=3, f_step=1, return_spectrum=False):
    """Quantify the rhythmicity of an oscillation using lagged coherence.

    Parameters
    ----------
    sig : 1d array
        Voltage time series.
    f_range : list of float
        Frequency range of the oscillation, as [low, high], in Hz.
    fs : float
        Sampling rate, in Hz.
    n_cycles : float, optional, default: 3
        Number of cycles of the frequency of interest used to compute lagged coherence.
    f_step : float, optional, default: 1
        Step size to calculate lagged coherence in the frequency range, in Hz.
    return_spectrum : bool, optional, default: False
        If True, return the lagged coherence for all frequency values. Otherwise, only return mean.

    Returns
    -------
    lc : float or 1d array
        If return_spectrum is False: mean lagged coherence value in the frequency range of interest.
        If return_spectrum is True: lagged coherence value for each frequency across the range.
    freqs : 1d array
        Frequencies, in Hz, corresponding to the lagged coherence values in lc.
        Only returned if return_spectrum is True.

    References
    ----------
    Fransen, A. M., van Ede, F., & Maris, E. (2015).
    Identifying neuronal oscillations using rhythmicity.
    Neuroimage, 118, 256-267.
    """

    # Identify Fourier components of interest
    freqs = np.arange(f_range[0], f_range[1] + f_step, f_step)

    # Calculate lagged coherence for each frequency
    lc = np.zeros(len(freqs))
    for ind, freq in enumerate(freqs):
        lc[ind] = _lagged_coherence_1freq(sig, freq, fs, n_cycles=n_cycles)

    # Return desired measure of lagged coherence
    if return_spectrum:
        return lc, freqs
    else:
        return np.mean(lc)


def _lagged_coherence_1freq(sig, freq, fs, n_cycles=3):
    """Calculate lagged coherence of sig at frequency freq using the hanning-taper FFT method"""

    # Determine number of samples to be used in each window to compute lagged coherence
    n_samps = int(np.ceil(n_cycles * fs / freq))

    # For each N-cycle chunk, calculate the fourier coefficient at the frequency of interest, freq
    chunks = _nonoverlapping_chunks(sig, n_samps)
    chunks_len = len(chunks)

    hann_window = hann(n_samps)
    fourier_f = np.fft.fftfreq(n_samps, 1 / float(fs))
    fourier_f_idx = np.argmin(np.abs(fourier_f - freq))
    fourier_coefsoi = np.zeros(chunks_len, dtype=complex)

    for ind, chunk in enumerate(chunks):
        fourier_coef = np.fft.fft(chunk * hann_window)
        fourier_coefsoi[ind] = fourier_coef[fourier_f_idx]

    # Compute the lagged coherence value
    lcs_num = 0
    for ind in range(chunks_len - 1):
        lcs_num += fourier_coefsoi[ind] * np.conj(fourier_coefsoi[ind + 1])
    lcs_denom = np.sqrt(np.sum(np.abs(fourier_coefsoi[:-1])**2) * np.sum(np.abs(fourier_coefsoi[1:])**2))

    return np.abs(lcs_num / lcs_denom)


def _nonoverlapping_chunks(sig, n_samples):
    """Split sig into nonoverlapping chunks of length N"""

    n_chunks = int(np.floor(len(sig) / float(n_samples)))
    chunks = np.reshape(sig[:int(n_chunks * n_samples)], (n_chunks, int(n_samples)))

    return chunks



"""The sliding window matching algorithm for identifying rhythmic components of a neural signal."""

import numpy as np

from neurodsp.utils.decorators import multidim

###################################################################################################
###################################################################################################

@multidim
def sliding_window_matching(sig, fs, win_len, win_spacing, max_iterations=500,
                            temperature=1, window_starts_custom=None):
    """Find recurring patterns in a time series using the sliding window matching algorithm.

    Parameters
    ----------
    sig : 1d array
        Voltage time series.
    fs : float
        Sampling rate, in Hz.
    win_len : float
        Window length, in seconds.
    win_spacing : float
        Minimum window spacing, in seconds.
    max_iterations : int
        Maximum number of iterations of potential changes in window placement.
    temperature : float
        Temperature parameter. Controls probability of accepting a new window.
    window_starts_custom : np.ndarray (1d), optional
        Pre-set locations of initial windows (instead of evenly spaced by 2G).

    Returns
    -------
    avg_window : 1d array
        The average waveform in 'sig' in the frequency 'f_range' triggered on 'trigger'.
    window_starts : 1d array
        Indices at which each window begins for the final set of windows.
    costs : 1d array
        Cost function value at each iteration.

    References
    ----------
    Gips, B., Bahramisharif, A., Lowet, E., Roberts, M. J., de Weerd, P.,
    Jensen, O., & van der Eerden, J. (2017). Discovering recurring
    patterns in electrophysiological recordings.
    Journal of Neuroscience Methods, 275, 66-79.
    MATLAB code: https://github.com/bartgips/SWM

    Notes
    -----
    * Apply a highpass filter if looking at high frequency activity,
      so that it does not converge on a low frequency motif.
    * win_len and win_spacing should be chosen to be about the size of the motif of
      interest, and the N derived should be about the number of occurrences.
    """

    # Compute window length and spacing in samples
    win_n_samps = int(win_len * fs)
    spacing_n_samps = int(win_spacing * fs)

    # Initialize window positions, separated by 2*G
    if window_starts_custom is None:
        window_starts = np.arange(0, len(sig) - win_n_samps, 2 * spacing_n_samps)
    else:
        window_starts = window_starts_custom
    n_windows = len(window_starts)

    # Randomly sample windows with replacement
    random_window_idx = np.random.choice(range(n_windows), size=max_iterations)

    # Calculate initial cost
    costs = np.zeros(max_iterations)
    costs[0] = _compute_cost(sig, window_starts, win_n_samps)

    for iter_num in range(1, max_iterations):

        # Pick a random window position to randomly replace with a
        # new window to improve cross-window similarity
        window_idx_replace = random_window_idx[iter_num]

        # Find a new allowed position for the window
        window_starts_temp = np.copy(window_starts)
        window_starts_temp[window_idx_replace] = _find_new_windowidx(
            window_starts, spacing_n_samps, len(sig) - win_n_samps)

        # Calculate the cost & the change in the cost function
        cost_temp = _compute_cost(sig, window_starts_temp, win_n_samps)
        delta_cost = cost_temp - costs[iter_num - 1]

        # Calculate the acceptance probability
        p_accept = np.exp(-delta_cost / float(temperature))

        # Accept update to J with a certain probability
        if np.random.rand() < p_accept:

            # Update costs & windows
            costs[iter_num] = cost_temp
            window_starts = window_starts_temp

        else:

            # Update costs
            costs[iter_num] = costs[iter_num - 1]

    # Calculate average window
    avg_window = np.zeros(win_n_samps)
    for w_ind in range(n_windows):
        avg_window = avg_window + sig[window_starts[w_ind]:window_starts[w_ind] + win_n_samps]
    avg_window = avg_window / float(n_windows)

    return avg_window, window_starts, costs


def _compute_cost(sig, window_starts, win_n_samps):
    """Compute the cost, which is proportional to the difference between pairs of windows"""

    # Get all windows and zscore them
    n_windows = len(window_starts)
    windows = np.zeros((n_windows, win_n_samps))

    for ind, window in enumerate(window_starts):
        temp = sig[window:window_starts[ind] + win_n_samps]
        windows[ind] = (temp - np.mean(temp)) / np.std(temp)

    # Calculate distances for all pairs of windows
    dists = []
    for ind1 in range(n_windows):
        for ind2 in range(ind1 + 1, n_windows):
            window_diff = windows[ind1] - windows[ind2]
            dist_temp = np.sum(window_diff**2) / float(win_n_samps)
            dists.append(dist_temp)

    # Calculate cost function, which is the average difference, roughly
    cost = np.sum(dists) / float(2 * (n_windows - 1))

    return cost


def _find_new_windowidx(window_starts, spacing_n_samps, n_samp, tries_limit=1000):
    """Find a new sample for the starting window"""

    for n_try in range(tries_limit):

        # Generate a random sample & check how close it is to other window starts
        new_samp = np.random.randint(n_samp)
        dists = np.abs(window_starts - new_samp)

        if np.min(dists) > spacing_n_samps:
            break

    else:
        raise RuntimeError('SWM algorithm has difficulty finding a new window. \
                            Increase the spacing parameter, G.')

    return new_samp



"""Time-frequency decompositions using wavelets."""

import numpy as np
from scipy.signal import morlet

from neurodsp.utils.decorators import multidim

###################################################################################################
###################################################################################################

@multidim
def morlet_transform(sig, fs, freqs, n_cycles=7, scaling=0.5):
    """Calculate the time-frequency representation of a signal using morlet wavelets.

    Parameters
    ----------
    sig : 1d array
        Time series.
    fs : float
        Sampling rate, in Hz.
    freqs : 1d array
        Frequency values to estimate with morlet wavelets.
    n_cycles : float
        Length of the filter, as the number of cycles of the oscillation
        whose frequency is the center of the bandpass filter.
    scaling : float
        Scaling factor.

    Returns
    -------
    mwt : 2d array
        Time-frequency representation of signal sig.
    """

    sig_len = len(sig)
    freqs_len = len(freqs)
    mwt = np.zeros([sig_len, freqs_len], dtype=complex)

    for f_ind, freq in enumerate(freqs):
        mwt[:, f_ind] = morlet_convolve(sig, fs, freq, n_cycles, scaling)

    return mwt


@multidim
def morlet_convolve(sig, fs, freq, n_cycles=7, scaling=0.5, filt_len=None, norm='sss'):
    """Convolve a signal with a complex wavelet.

    Parameters
    ----------
    sig : 1d array
        Time series to filter.
    fs : float
        Sampling rate, in Hz.
    freq : float
        Center frequency of bandpass filter.
    n_cycles : float, optional, default=7
        Length of the filter, as the number of cycles of the oscillation with specified frequency.
    scaling : float, optional, default=0.5
        Scaling factor for the morlet wavelet.
    filt_len : integer, optional
        Length of the filter. If not None, this overrides the freq and n_cycles inputs.
    norm : {'sss', 'amp'}, optional
        Normalization method:

        * 'sss' - divide by the sqrt of the sum of squares of points
        * 'amp' - divide by the sum of amplitudes divided by 2

    Returns
    -------
    array
        Complex time series.

    Notes
    -----

    * The real part of the returned array is the filtered signal.
    * Taking np.abs() of output gives the analytic amplitude.
    * Taking np.angle() of output gives the analytic phase.
    """

    if n_cycles <= 0:
        raise ValueError('Number of cycles for morlet wavelets must be a positive number.')

    if filt_len is None:
        filt_len = n_cycles * fs / freq

    morlet_f = morlet(filt_len, w=n_cycles, s=scaling)

    # ToDo: there is an inconsistency between these methods, and the docs.
    if norm == 'sss':
        morlet_f = morlet_f / np.sqrt(np.sum(np.abs(morlet_f)**2))
    elif norm == 'abs':
        morlet_f = morlet_f / np.sum(np.abs(morlet_f))
    else:
        raise ValueError('Not a valid wavelet normalization method.')

    mwt_real = np.convolve(sig, np.real(morlet_f), mode='same')
    mwt_imag = np.convolve(sig, np.imag(morlet_f), mode='same')

    return mwt_real + 1j * mwt_imag
